from util import inf

class IndexingDfunc(object):
    def __init__(self,dfunc,maxpivots=20,autoadd=True):
        self.dfunc = dfunc
        self.pointlist = {}
        self.maxpivots = maxpivots
        self.autoadd = True
        self.query = None
    
    def __call__(self,a,b):
        if not self.query:
            dist = self.dfunc(a,b)
            if self.autoadd:
                self.addPoint(a,b,dist)
            return dist
        else:
            return self.query(a,b)
            
    
    def addPoint(self,obj,piv,dist):
        if not self.pointlist.has_key(obj):
            self.pointlist[obj] = IndexDist(obj,self.maxpivots,self)
        if not self.pointlist.has_key(piv):
            self.pointlist[piv] = IndexDist(piv,self.maxpivots,self)
        self.pointlist[obj].addDist(piv,dist)
    
    def startquery(self,query):
        self.query = QueryDfunc(self,query)
    
    def stopquery(self):
        self.query = None
        

class QueryDfunc(object):
    def __init__(self,indfunc,query):
        self.indfunc = indfunc
        self.query = query
        self.approxdist = {}
    
    def __call__(self,a,b):
        if a == self.query:
            return self.retappdist(b)
        elif b == self.query:
            return self.retappdist(a)
        else:
            raise Exception("Query object not present (%s,%s)"%(str(a),str(b)))
            
    def retappdist(self,a):
        #print a
        if not self.approxdist.has_key(a):
            self.approxdist[a] = ApproximatePoint(self.indfunc.pointlist[a],self)
        return self.approxdist[a]

class IndexDist(object):
    def __init__(self,o,maxpivots,indfunc):
        self.o = o
        self.dlist = []
        self.maxpivots = maxpivots
        self.indfunc = indfunc
        self.pivobs = 0
        
        
    def compute(self,query):
        return self.indfunc.dfunc(query,self.o)
        
    def addDist(self,pivot,dist):
        #We choose the pivots that have the shortest distance to the object
        if len(self.dlist) < self.maxpivots and not (dist,pivot) in self.dlist:
            self.dlist.append((dist,pivot))
            self.pivobs += 1
        else:
            maxobj = max(self.dlist)
            #print self.dlist,maxobj
            if maxobj > (dist,pivot):
                self.dlist.remove(maxobj)
                self.dlist.append((dist,pivot))
                

class ApproximateDist(object):
    def __init__(self):
        self.dirty = True
        self.minimum = 0
        self.maximum = inf
       #print self.__class__
        self.computed = False
    
    def getMinimum(self):
        return self.minimum
        
    def getMaximum(self):
        return self.maximum
    
    #Tries to improve the accuracy of the object
    def improve(self):
        raise NotImplementedError

    def update(self):
        pass
        
    def __str__(self):
        return "[%f,%f]"%(str(self.getMinimum()),self.getMaximum())
    def betterobj(self,otherobj):
        return self.getMaximum()-self.getMinimum() > otherobj.getMaximum()-otherobj.getMinimum() and self or otherobj
    
    def __add__(self,otherobj):
        return ApproximateAdd(self,otherobj)
    
    def __radd__(self,otherobj):
        return self.__add__(otherobj)
    
    def __sub__(self,otherobj):
        return ApproximateSub(self,otherobj)
    
    def __rsub__(self,otherobj):
        return ApproximateSub(otherobj,self)
    
    def __mul__(self,otherobj):
        return ApproximateMul(self,otherobj)
    
    def __rmul__(self,otherobj):
        return self.__mul__(otherobj)

    def __cmp__(self,otherobj):
        self.update()
        #print "Cmp"
        #print self,otherobj
        if isinstance(otherobj,ApproximateDist):
            while self.getMaximum() > otherobj.getMinimum() and self.getMinimum() < otherobj.getMaximum():
                self.betterobj(otherobj).improve()
            if self.getMaximum()< otherobj.getMinimum():
                return -1
            elif self.getMinimum() > otherobj.getMaximum():
                return 1
            else:
                return 0
        else: #Assume it is possible to compare with
            #print self.getMaximum(),self.getMinimum(),otherobj
            while self.getMaximum() > otherobj and self.getMinimum() < otherobj:
                self.improve()
            if self.getMaximum() < otherobj:
                return -1
            elif self.getMinimum() > otherobj:
                return 1
            else:
                return 0    
                                        
class ApproximatePoint(ApproximateDist):
    def __init__(self,indexdist,querydfunc):
        super(ApproximatePoint,self).__init__()
        self.indexdist = indexdist
        self.querydfunc = querydfunc
            
    def improve(self):
        if self.computed:
            #print "Improve an already computed"
            return
        dist = self.indexdist.compute(self.querydfunc.query)
        self.minimum = dist
        self.maximum = dist
        self.computed = True
        
    def update(self):
        oldmin = self.minimum
        oldmax = self.maximum
        ncount = 0
        for d,pivot in self.indexdist.dlist:
            appivot = self.querydfunc.retappdist(pivot)
            #appivot.improve()
            #if appivot.getMinimum() > 0:
            #    print appivot
            if appivot.minimum == appivot.maximum:
                ncount += 1
            self.minimum = max(self.minimum,appivot.minimum-d)
            self.maximum = min(self.maximum,appivot.maximum+d)
        
        #print ncount,len(self.indexdist.dlist)
class ApproximateConst(ApproximateDist):
    def __init__(self,val):
        super(ApproximateConst,self).__init__()
        self.val = val
        self.computed = True
        #print dir(self)
    
    def getMinimum(self):
        return self.val
    
    def getMaximum(self):
        return self.val
    
    def improve(self):
        print "Improve on const",self.val
        
    
    def __str__(self):
        return "%fC"%self.val
#An operation that adds, subtracts or multiplicates
class ApproximateOp(ApproximateDist):
    def __init__(self,left,right):
        super(ApproximateOp,self).__init__()
        self.left = self.checkApp(left)
        self.right = self.checkApp(right)
    
    def checkApp(self,val):
        if not isinstance(val,ApproximateDist):
            return ApproximateConst(val)
        else:
            return val
    
    def improve(self):
        self.left.betterobj(self.right).improve()
        if self.left.computed and self.right.computed:
            self.computed = True
    
    def update(self):
        self.left.update()
        self.right.update()
        
    def __str__(self):
        return "%s(%s,%s)"%(str(self.__class__),str(self.left),str(self.right))

class ApproximateAdd(ApproximateOp):
    def getMinimum(self):
        return self.left.getMinimum()+self.right.getMinimum()
    def getMaximum(self):
        return self.left.getMaximum()+self.right.getMaximum()
    
class ApproximateSub(ApproximateOp):    
    def getMinimum(self):
        return self.left.getMinimum()-self.right.getMaximum()
    
    def getMaximum(self):
        return self.left.getMaximum()-self.right.getMinimum()
    
class ApproximateMul(ApproximateOp):
    
    def getMinimum(self):

        return min(self.getAllVals())
    def getMaximum(self):

        return max(self.getAllVals())
            
    def getAllVals(self):
        return [self.left.getMinimum()*self.right.getMinimum(),
                self.left.getMinimum()*self.right.getMaximum(),
                self.left.getMaximum()*self.right.getMinimum(),
                self.left.getMaximum()*self.right.getMaximum()]
