"""
The BS-Tree index structure.
"""

from util import DistQueue, bunch, ball_overlap
from sets import Set as set

__metaclass__ = type

class BSTree:

    pivot_count = 2
    
    def __init__(self, metric):
        self.data = metric
        self.dfunc = metric.getDistance
        self.root = bunch(pivots=[],
                          rads=[0]*self.pivot_count,
                          children=[None]*self.pivot_count)
        for obj in self.data:
            self.add(obj)
    
    # For the Voronoi tree
    def new_child_hook(self, parent, child, idx):
        pass

    def add(self, obj):
        d = self.dfunc
        def add(node):
            if len(node.pivots) < self.pivot_count:
                node.pivots.append(obj)
            else:
                dp = [d(obj, p) for p in node.pivots]
                idx = 0
                for i in range(len(dp)):
                    if dp[i] < dp[idx]:
                        idx = i
                child = node.children[idx]
                if child is None:
                    child = bunch(pivots=[],
                                  rads=[0]*self.pivot_count,
                                  children=[None]*self.pivot_count)
                    node.children[idx] = child
                    self.new_child_hook(node, child, idx)
                if dp[idx] > node.rads[idx]:
                    node.rads[idx] = dp[idx]
                add(child)
        add(self.root)

    def range_search(self, query, radius):
        result = []
        # In the Voronoi tree, there may be duplicates in the result. We use a 
        # separate set of IDs to prevent this, as the objects themselves may 
        # not be hashable or even properly comparable (c.f. NumPy arrays).
        ids = set()
        def search(node):
            for p, c, r in zip(node.pivots, node.children, node.rads):
                dist = self.dfunc(query, p)
                if dist <= radius and not id(p) in ids:
                    result.append(p)
                    ids.add(id(p))
                if c is None: continue
                if ball_overlap(radius, r, dist):
                    search(c)
        search(self.root)
        return result
    
    def knn_search(self, query, count, range=None):
        distance = self.dfunc
        queue = DistQueue(query, count, distance)
        def search(node):
            dups = node.get('dups')
            for p, c, r in zip(node.pivots, node.children, node.rads):
                dist = self.dfunc(query, p)
                # Used in the Voronoi tree, which contains duplicates
                # in descendant lines to the leaves:
                # (Note that id(p) is used because the objects may be 
                # unhashable.)
                if not dups or not id(p) in dups:
                    queue.add(p)
                if c is None: continue
                if not queue.is_full() or \
                        ball_overlap(queue.get_radius(), r, dist):
                    search(c)
        search(self.root)
        return queue.get_contents(range)

USABLE = True
INDEX_NAME = "BS-Tree"
INDEX_CLASS = BSTree
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Kalantari:1983"
