"""
The BU-Tree index structure, which is simply a BS-Tree built bottom-up, using
an agglomerative clustering procedure.

http://www.springerlink.com/content/12262h6134632w44/

"""
import util
from util import bunch

__metaclass__ = type

class BUTree:
    
    def __init__(self, data, dfunc):
        self.data = list(data)
        self.n = len(self.data)
        self.dfunc = dfunc
        self.build_tree()
    
    def build_tree(self):
        n = self.n
        data = self.data
        nodes = {}
        for objid in range(n):
            nodes[objid] = bunch(
                pivot = objid,
                radius = 0,
                left = None,
                right = None
            )
        matrix = util.DMArray((n,n))
        for i in range(n):
            for j in range(n):
                if i == j:
                    matrix[i,j] = util.inf
                else:
                    matrix[i,j] = self.dfunc(data[i],data[j])
        while len(nodes) > 1:
            minval = util.inf
            minkey = None
            for key, val in matrix.iteritems():
                if val < minval:
                    minval = val
                    minkey = key
            i, j = minkey
            node1 = nodes[i]
            node2 = nodes[j]
            p = bunch(
                pivot = i,
                radius = matrix[i,j],
                left = node1,
                right = node2
            )
            nodes[i] = p
            del nodes[j]
            for k in range(n):
                matrix[k,i] = max(matrix[k,i], matrix[k,j])
            for k in range(n):
                matrix[k,j] = matrix[j,k] = util.inf
        self.root = nodes.popitem()[1]
    
    def range_search(self, query, range):
        result = {}
        def search(node):
            if node is None:
                return
            pivot = self.data[node.pivot]
            dist = self.dfunc(pivot, query)
            if dist <= node.radius + range:
                if dist <= range:
                    # Avoid duplicates, without assuming hashability:
                    result[id(pivot)] = pivot
                search(node.left)
                search(node.right)
        search(self.root)
        return [o for (k, o) in result.iteritems()]

USABLE = False
INDEX_NAME = "BU-Tree"
INDEX_CLASS = BUTree
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Liu:2006"
SQUARED_MEM = True
