"""
The D-Index index structure.
"""
from __future__ import generators
import util
import pivotselector
from laesa import LAESA
from linearscan import LinearScan
import math
from util import Bucket

__metaclass__ = type

def psplit(dist,p,r):
    if dist <= r - p:
        return 0
    elif dist > r + p:
        return 1
    else:
        return -1


class DIndex:

    def __init__(self,metric,numpivots=5,levels=5,p=0.1,bestDivision=None,hybrid=True,pselector=pivotselector.BustosSelector,oldpspace=None,buckets=[]):
        
        self.pselector = pselector
        self.data = metric
        #print numpivots,levels,p,len(self.data)
        self.dfunc = metric.getDistance
        self.numpivots = numpivots
        self.levels = levels
        self.p = p
        self.bestDivision = bestDivision
        if oldpspace == None:
            oldpspace = [list() for x in self.data]
        self.index(oldpspace)
        self.hybrid = hybrid
        
        
        
    def index(self,oldpspace):
        
        #Choose n objects as pivots (random)
        #self.pivots = util.sample(self.data,self.numpivots)
        self.pivots = self.pselector(self.data,self.dfunc).generate(self.numpivots)
        
        
        #Find median distance for each pivot
        self.medians = []
        self.dists = []
        exclusion = []
        for pivot in self.pivots:
            ldists = []
            for object in self.data:
                ldists.append(self.dfunc(object,pivot))
            lsorted = list(ldists)
            util.sort(lsorted)
            if self.bestDivision != None:
                med = util.bestDivision(lsorted,self.p,self.bestDivision)
            else:
                med = lsorted[len(lsorted)/2]
            
            #print med,lsorted[len(lsorted)/2]
            self.medians.append(med)
            self.dists.append(ldists)
        buckets = {}
            
        npspace = []
        for po,object in enumerate(self.data):
            #First calculate hash value
            if object in self.pivots:
                continue
            m = 1
            hash = 0
            for pp,pivot in enumerate(self.pivots):
                pdist = psplit(self.dists[pp][po],self.p,self.medians[pp])
                
                if pdist == -1:
                    hash = pow(2,len(self.pivots))
                    break
                hash += m*pdist
                m*=2
            #print hash
            
            opspace = list(oldpspace[po])
            opspace.extend([self.dists[i][po] for i in xrange(len(self.pivots))])
            
            if hash == pow(2,len(self.pivots)):
                exclusion.append(object)
                npspace.append(opspace)
            else:
                if not buckets.has_key(hash):
                    buckets[hash] = []
                buckets[hash].append((object,opspace))
        
        self.buckets = {}
        for key,val in buckets.items():
            #We use log(len(val)) pivot values for LAESA, should be fairly ok
            #self.buckets[key] = LinearScan(util.GenericMetricList(val))
            self.buckets[key] = Bucket(self.dfunc,val)
        
        #print "Level " + str(self.levels)
        #print self.buckets
        
        
        
        #print len(exclusion)
        if len(exclusion) > 0:
            if self.levels == 1:
                ilist = []
                for obj,space in zip(exclusion,npspace):
                    ilist.append((obj,space))
                self.exclusion = Bucket(self.dfunc,ilist)
                print len(exclusion)
            else:
                self.exclusion = DIndex(util.GenericMetricList(exclusion),self.numpivots-1,self.levels-1,self.p*(float((self.levels-1))/self.levels),bestDivision=self.bestDivision,pselector=self.pselector,oldpspace=npspace)
        else:
            self.exclusion = None
        
  
  
  
    #Assumes that the # of pivots > 0 (#pivots = 0 should not happen)
    def findAllHashes(self,query,hash=None,p=None):
        if not hash:
            hash = self.getHashArray(query,p)
        hashlist = [0]
        m = 1
        for val in hash:
            if val < 0:
                hashlist.extend([x+m for x in hashlist])
            elif val == 1:
                hashlist = [x+m for x in hashlist]
            m*=2
        return hashlist
                
            
            
        
  
    def findhash(self,query=None,p=None,hash=None):
        retval = 0
        m = 1
        if not hash:
            hash = self.getHashArray(query,p,abort=True)
        for val in hash:
            if val < 0:
                return pow(2,len(hash))
            retval += m*val
            m *=2
        return retval
        """
        if p == None:
            p = self.p
        print p
        hash = 0
        m = 1
        for pp,pivot in enumerate(self.pivots):
            pdist = psplit(self.dfunc(query,pivot),p,self.medians[pp])
            #print pdist
            
            if pdist < 0:
                return pow(2,len(self.pivots))
            hash += m*pdist
            m *= 2
        return hash
        """
    
    def getHashArray(self,query,p=None,abort=False,distarr=None):
        if p == None:
            p = self.p
        hash = [-1] * len(self.pivots)
        i = 0
        if distarr == None:
            distarr = getDistArray(query,self.pivots)
        for dist,med in zip(distarr,self.medians):
            pdist = psplit(dist,p,med)
            if abort and pdist < 0:
                break
            hash[i] = pdist
            i += 1
        return hash


    def getDistArray(self,query,pivlist):
        for pivot in pivlist:
            yield self.dfunc(query,pivot)
    
    def range_search(self,query,radius,pspace=None):
        #print self,query,self.levels
        retlist = list()
        if not pspace:
            pspace = []
        #print pspace
            
        
        #print self.levels
        #print self.levels
        rho = max(0,radius-self.p)
        #print self.findAllHashes(query,rho)
        distarr = list(self.getDistArray(query,self.pivots))
       
        for dist,object in zip(distarr,self.pivots):
            if dist < radius:
                retlist.append(object)
        pspace.extend(distarr)
        #print distarr
        hasharr = self.getHashArray(query,rho,False,distarr)
        hashes = self.findAllHashes
        for hash in self.findAllHashes(query,hasharr):
            
            #hash = self.findhash(query,0)
            #print hash
            if self.buckets.has_key(hash):
                #print self.p,self.levels,hash,len(self.buckets[hash].objects),len(pspace),pspace
                tlist = self.buckets[hash].range_search(query,radius,pspace)
                [retlist.append(o) for o in tlist]
            #print hash,self.buckets[hash]
            
        if self.exclusion:
            #print "Exclusion"
            [retlist.append(o) for o in self.exclusion.range_search(query,radius,list(pspace))]
        #print [str(r) for r in retlist]
        #print len(retlist)
        return retlist
    

    
    def optimistic_knn_search(self,query,n,queue):
        dk = len(queue) < n and util.inf or queue.get_radius()
        radius = min(self.p,dk)
        sepbucket = findhash(query,self.p+radius)
        if sepbucket < pow(2,len(self.pivots)):
            #Exclusive containment in separable bucket
            if self.buckets.has_key(sepbucket):
                for o in self.buckets[sepbucket].knn_search(n):
                    queue.add(o)
                if len(queue) == n and queue.get_radius() < self.p:
                    return
        
        else:
            expbucket = getHashArray(query,max(0,self.p-radius))
            if expbucket < pow(2,len(self.pivots)):
                if self.buckets.has_key(expbucket):
                    for o in self.buckets[expbucket].knn_search(n):
                        queue.add(o)
        if self.levels > 1:
            self.optimistic_knn_search(query,n,queue)
                    
        
        
        
    
    def unoptimistic_knn_search(self,query,n,queue):
        pass
    
    def knn_search(self,query,n):
        #First, be optimistic
        queue = util.DistQueue(n);
        
        retlist = self.optimistic_knn_search(query,n,queue)
        if len(retlist) < n:
            print "Unoptimistic search"
            #retlist.append(self.unoptimistic_knn_search(query,n,retlist)))
        return retlist
                                                   
        

USABLE = True
INDEX_NAME = "D-Index"
INDEX_CLASS = DIndex
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Dohnal:2001,Dohnal:2003,Dohnal:2004,Dohnal:2004a"