"""
The EM-VP-Forest index structure.

Excluded Middle Vantage Point Forest

http://citeseer.ist.psu.edu/368813.html

"""

__metaclass__ = type

from util import GenericMetricList
from aesa import AESA
from util import Bucket
import random
from sets import Set
#from sets import Set as set
import util

class EMVPForest:
    def __init__(self,metric,maxrange=0.2,bucketsize=40,maxpivots=5,trees=5):
        self.forest = []
        middle = list(metric)
        self.tau = maxrange
        tree = 10
        while len(middle) > 0:
            #print len(middle),self.tau
            #print middle
            nmiddle = []
            emvpt = EMVPTree([(o,[]) for o in middle] ,metric.getDistance,self.tau*(float(tree)/trees),bucketsize,nmiddle,maxpivots)
            self.forest.append(emvpt)
            middle = nmiddle
            tree-= 1
        #self.forest.append(AESA(GenericMetricList(middle)))
    
    def range_search(self,query,range):
        retlist = []
        for tree in self.forest:
            #print retlist
            retlist.extend(tree.range_search(query,range))
        return retlist




class EMVPTree:
    def __init__(self,data,dfunc,tau,bucketsize,middlelist,maxpivots):
        self.dfunc = dfunc
        self.tau = tau
        self.bucketsize = bucketsize
        self.middlelist = middlelist
        self.maxpivots = maxpivots
        self.index(data)
    
    def divide(self,indata):
        #Assumse len(indata) >= 1
        self.center = indata[random.randint(0,len(indata)-1)][0]
        
        inside = []
        outside = []
        #print indata
        #dlist =[]
        #for o,ps in indata:
        #    dlist.append(self.dfunc(self.center,o))
        dlist = [self.dfunc(self.center,o) for o,ps in indata ]
        
        #print len(dlist),len(bf),dlist
        dclist = list(dlist)
        dclist.sort()
        #if len(bf) != len(af) or len(Set(dlist)) != len(Set(dclist)):
        #    print len(dlist),len(af),dclist
        #    print dlist
        #    print af.difference(bf),bf.difference(af)
        self.median = dclist[len(dclist)/2]
        for d,(o,ps) in zip(dlist,indata):
            ps.append(d)
            if d == 0:
                #It is the center
                pass
            elif d <= self.median-self.tau:
                inside.append((o,ps)) 
            elif d > self.median+self.tau:
                outside.append((o,ps))
            else:
                self.middlelist.append(o)
        return (inside,outside)
        
    def index(self,data):
        if len(data) <= self.bucketsize:
            self.inside = util.Bucket(self.dfunc,data)
            self.outside = util.Bucket(self.dfunc,[])
        (inside,outside) = self.divide(data)
        #print len(inside),len(outside),len(self.middlelist)
        #print "Center",self.center
        #print "Ins",inside
        #print "Out",outside
        if len(inside) > self.bucketsize:
            self.inside = EMVPTree(inside,self.dfunc,self.tau,self.bucketsize,self.middlelist,self.maxpivots)
        else:
            self.inside = util.Bucket(self.dfunc,inside)
        if len(outside) > self.bucketsize:
            self.outside = EMVPTree(outside,self.dfunc,self.tau,self.bucketsize,self.middlelist,self.maxpivots)
        else:
            self.outside = util.Bucket(self.dfunc,outside)
    
    def range_search(self,query,range,qspace=None):
        retlist = []
        if not qspace:
            qspace = list()
        dp = self.dfunc(query,self.center)
        qspace.append(dp)
        if dp < range:
            retlist.append(self.center)
        if dp <= self.median+range-self.tau:
            retlist.extend(self.inside.range_search(query,range,list(qspace)))
        if dp >= self.median-range+self.tau:
            retlist.extend(self.outside.range_search(query,range,list(qspace)))
        return retlist
                
        
            

'''

class EMVPTree:
    def __init__(self,dfunc,tau,bucketsize):
        self.dfunc = dfunc
        self.tau = tau
        self.bucketsize = bucketsize

        
    def index(self,data):
        self.center = data[0]
        if self.center in data[1:]:
            print "Center",self.center
        if len(data) ==1:
            self.right = None
            self.left = None
            self.median = 0
            return []
        #print self.center
        dists = [(self.dfunc(self.center,o),o) for o in data if not o == self.center]
        dists.sort()
        
        self.median = dists[len(dists)/2][0]
        
        lefts = [o for d,o in dists if d <= self.median-self.tau]
        rights = [o for d,o in dists if d >= self.median+self.tau]
        bucket = [o for d,o in dists if d > self.median-self.tau and d < self.median+self.tau]
        
        print self.median-self.tau,self.median+self.tau,self,self.center,len(lefts),len(rights),len(bucket)
        #print lefts,rights,bucket
        lbucket = self.buildindex(list(lefts),"left")
       
        #print self.median-self.tau,self.median+self.tau,self,self.center,len(lefts),len(rights),len(bucket)
        rbucket = self.buildindex(list(rights),"right")
        #print self.median-self.tau,self.median+self.tau,self,self.center,len(lefts),len(rights),len(bucket)
        #print lefts,rights,bucket
        
        for b in bucket:
            if b in lbucket:
                print "Left",b
            if b in rbucket:
                print "Right",b
                
        bucket.extend(lbucket)
        bucket.extend(rbucket)
        return bucket
    
    
    
    def buildindex(self,ilist,varname):
        if len(ilist) < self.bucketsize:
            setattr(self,varname,AESA(GenericMetricList(ilist)))
            #print ilist
            return []
        else:
            emvpt = EMVPTree(self.dfunc,self.tau,self.bucketsize)
            setattr(self,varname,emvpt)
            return emvpt.index(ilist)
    
    def range_search(self,query,range):
        #Assume range < tau
        retlist = []
        dp = self.dfunc(query,self.center)
        #print self.center,dp
        
        #print retlist
        if dp <= self.median and not self.left == None:
            retlist.extend(self.left.range_search(query,range))
        #print retlist
        if dp >= self.median and not self.right == None:
            retlist.extend(self.right.range_search(query,range))
        #print retlist
        
        if self.center in retlist:
            print "!!!!!!!!!!!!!!!!!"
            print self.center
            print str(self.left) + str(hasattr(self.left,'data') and self.left.data or "")
            print str(self.right) + str(hasattr(self.right,'data') and self.right.data or "")
        
        if dp < range:
            retlist.append(self.center)
        return retlist

'''


USABLE = True
INDEX_NAME = "EM-VP-Forest"
INDEX_CLASS = EMVPForest
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Yianilos:1999"
