"""
The GH-Tree index structure.
"""

from __future__ import division
from util import bunch

__metaclass__ = type

class GHTree:
    
    def __init__(self, metric):
        self.data = metric
        self.dfunc = metric.getDistance
        self.root = self.new_node()
        for obj in metric:
            self.add(obj)
    
    def new_node(self):
        return bunch(
            pleft=None, pright=None,
            left=None, right=None)
    
    def add(self, obj):
        def add(node):
            if node.pleft is None:
                node.pleft = obj
                node.left = self.new_node()
            elif node.pright is None:
                node.pright = obj
                node.right = self.new_node()
            else:
                dleft = self.dfunc(obj, node.pleft)
                dright = self.dfunc(obj, node.pright)
                if dleft < dright:
                    add(node.left)
                else:
                    add(node.right)
        add(self.root)
    
    def range_search(self, query, radius):
        result = []
        def search(node):
            if node.pleft is not None:
                dleft = self.dfunc(query, node.pleft)
                if dleft <= radius:
                    result.append(node.pleft)
            if node.pright is not None:
                dright = self.dfunc(query, node.pright)
                if dright <= radius:
                    result.append(node.pright)
            if node.pleft is None or node.pright is None:
                return
            if (dleft - dright)/2 <= radius:
                search(node.left)
            if (dright - dleft)/2 <= radius:
                search(node.right)
        search(self.root)
        return result

USABLE = True
INDEX_NAME = "GH-Tree"
INDEX_CLASS = GHTree
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Uhlmann:1991,Uhlmann:1991a"