"""
The GNAT index structure.
"""

__metaclass__ = type

import pivotselector
import util
from aesa import AESA



class GNAT:
    def __init__(self,metric,k=10,maxbucket=0,pivsel=pivotselector.RandomSelector,bucketmethod=AESA):
        global memusage
        memusage = 0
        self.root = GNATNode(metric,metric.getDistance,k,pivsel,bucketmethod,maxbucket)
        print k,memusage
    def range_search(self,query,range):
        return self.root.range_search(query,range)
    

class GNATNode:
    def __init__(self,list,dfunc,k,pivsel,bucketmethod,maxbucket):
        self.dfunc = dfunc
        self.pivsel = pivsel
        self.k = k
        self.maxbucket = maxbucket
        self.bucketmethod = bucketmethod
        self.index(list)
        
    def index(self,list):
        global memusage

        self.pivots = self.pivsel(list,self.dfunc).generate(self.k)
        memusage += len(self.pivots)*len(self.pivots)
        
        self.rangearr = [ [(util.inf,0)]*len(self.pivots) for p in self.pivots]
        
        pivdata = [[] for p in self.pivots]
        
        for obj in list:
            dists = [self.dfunc(pivot,obj) for pivot in self.pivots]
            minpiv = min([(n,i) for (i,n) in enumerate(dists)])[1]
            for i,pdist in enumerate(dists):
                self.rangearr[i][minpiv]= (min(self.rangearr[i][minpiv][0],pdist),max(self.rangearr[i][minpiv][1],pdist))
            if not obj in self.pivots:
                pivdata[minpiv].append(obj)
        
        avgdata = len(list)/self.k    
        
        self.pivnodes= [None]*len(self.pivots)
        for i,pivot in enumerate(self.pivots):
            if len(pivdata[i]) == 0:
                continue
            if len(pivdata[i]) <= self.maxbucket:
                #print "AESA"
                
                self.pivnodes[i] = self.bucketmethod(util.GenericMetricList(pivdata[i]))
            else:    
                #pk = min(200,max(2,(self.k*len(pivdata[i]))/avgdata))
                pk = self.k
                #print pk
                self.pivnodes[i] = GNATNode(pivdata[i],self.dfunc,pk,self.pivsel,self.bucketmethod,self.maxbucket)
            

    def range_search(self,query,radius):
        #while we still have points left, remove poi    
        P = range(len(self.pivots))
        retlist = []
        
        for l in range(len(self.pivots)):
            if not l in P:
                continue
            qr = self.dfunc(self.pivots[l],query)
            if qr < radius:
                retlist.append(self.pivots[l])
            nP = []
            for p in P:
                (dmin,dmax) = self.rangearr[l][p]
                #print (p,dmin,dmax)
                if qr-radius <= dmax and qr+radius >= dmin:
                    nP.append(p)
            #print nP
            P = nP

        for p in P:
            if self.pivnodes[p]:
                retlist.extend(self.pivnodes[p].range_search(query,radius))
        return retlist

USABLE = True    
INDEX_NAME = "GNAT"
INDEX_CLASS = GNAT
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Brin:1995"