"""
The HC index structure.
"""

# XXX See LC. Code sharing should be possible. (I.e., LC could be a subclass 
# with limited functionality)
# Differences:
# 1. Build a similar structure recursively for the inside-set.
# 2. Use a h-function (h(n)) to control the bucket size for a node containing 
# n objects.
# Could use use a lower limit to bucket size (i.e., if there are at most b 
# objects, don't recurse) to save space -- but that could increase the number 
# of distance comparisons.

# XXX See the experiments by Andersson to figure out the optimal parameters,
# as well as the choice of h(n).

# Each node has a covering radius for the union of the inside and outside
# parts -- so that if d(c,q) > cr'(c) + r, the recursion can stop. This can 
# also be used for automatic inclusion, both for the inside and outside
# parts.

# Fredriksson uses a different pivot selection method from Chavez and Navarro;
# he minimizes the mean squared distance to the rest, using random sampling.

# In Fredriksson:2005 this is called UVPT (unbalanced VP-tree)
# XXX (Or is this really the same?)

__metaclass__ = type



from random import randint
from util import GenericMetricList
from linearscan import LinearScan

class HC:
    def __init__(self,metric):
        self.bucketsize = 20
        data = metric
        self.dfunc = metric.getDistance
        self.index(data)
    
    def bucket(self,glist):
        return LinearScan(glist)

    def split(self,list):
        n = len(list)/2
        return n
    
    def indexlist(self,data):
        if hasattr(self,'centerobject') and self.centerobject in data:
            data.remove(self.centerobject)
        glist = GenericMetricList([o for (d,o) in data])
        if len(data) < self.bucketsize or self.split(data) == 0:
            return self.bucket(glist)
        else:
            return self.__class__(glist)
    
    def indexinc(self,inclist):
        self.included = self.indexlist(inclist)
        
    def indexenc(self,exclist):
        self.excluded  = self.indexlist(exclist)
    
    def index(self,data):
        #Take random center object        
        self.centerobject = data[randint(0,len(data)-1)]
        dlist = [(self.dfunc(self.centerobject,o),o) for o in data if o is not self.centerobject]
        self.rcmax = max(dlist)[0]
        dlist.sort()
        hn = self.split(dlist)
        #print hn,len(dlist)
        (excluded,included) = (dlist[hn:],dlist[:hn])
        self.incrc = max(included)[0]
        self.indexinc(included)
        self.indexenc(excluded)
    
    def range_search(self,query,range):
        retlist = []
        dc = self.dfunc(query,self.centerobject)
        if dc < range:
            retlist.append(self.centerobject)
        if dc > self.incrc-range:

            retlist.extend(self.excluded.range_search(query,range))
        if dc < self.incrc+range:
            retlist.extend(self.included.range_search(query,range))
        return retlist
                               
            
            

USABLE = True
INDEX_NAME = "HC"
INDEX_CLASS = HC
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Fredriksson:2005,Fredriksson:2006"