"""
The iAESA index structure.
"""

__metaclass__ = type

from aesa import AESA

class IAESA(AESA):
    
    def set_up_heuristic(self):
        self.perm = {}
    
    # XXX A bit slow, but not the slowest part...
    def pivot_seen_hook(self, u, p, d):
        """
        Adds p to the pivot permutation of u.
        """
        perm = self.perm.setdefault(u, [])
        perm.append((d, p))
        perm.sort() # Simpler than bisecting and inserting or whatnot
    
    # XXX Super-slow!
    def heuristic(self, q, u, p, old):
        """
        Instead of the sum of lower bounds used in AESA, iAESA uses the correlation
        (in the form of Spearman's Footrule) between distance-sorted pivot
        permutations for q and u.
        """
        # XXX Doesn't use the F() from util, because that works with a different
        # permutation format. Should perhaps be consolidated...?
        def perm(x):
            return sorted([(o, i) for (i, o) in enumerate(self.perm[q])])
        return sum([abs(a[1]-b[1]) for (a, b) in zip(perm(q), perm(u))])

USABLE = False
INDEX_NAME = "iAESA"
INDEX_CLASS = IAESA
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Figueroa:2006"
SQUARED_MEM = True
