package bonsai.indexes;

import java.util.Arrays;

public class LazySlimTreeNode<T> extends SlimTreeNode<T> {

	private boolean squaredTime;
	
	public LazySlimTreeNode(DistanceFunction df, int capacity, boolean squaredTime){
		super(df, capacity);
		this.squaredTime = squaredTime;
	}
	
	public LazySlimTreeNode(DistanceFunction df, int capacity, MTreeNodeType type, boolean squaredTime){
		super(df, capacity, type);
		this.squaredTime = squaredTime;
	}

	@Override
	protected AbstractMTreeNode<T> createNewNode(MTreeNodeType type) {
		return new LazySlimTreeNode<T>(df, capacity, type, squaredTime);
	}

	@Override
	protected double[][] getMST(int[] addFrom, double[] addCosts) {
		int n = entries.size();
		
		int pivot = -1;
		for(int i = 0; i < n; i++) if(entries.get(i).obj == getPivot()) pivot = i;
		double[][] dists = new double[n][n];
		
		for(int i = 0; i < n; i++){
			Arrays.fill(dists[i], -1);
			dists[i][i] = 0;
		}
		for(int i = 0; i < n; i++){
			dists[i][pivot] = dists[pivot][i] = ((MTreeEntry<T>)entries.get(i)).pivotDistance;
		}
		boolean[] included = new boolean[n];
		included[pivot] = true;
		for(int i = 0; i < n; i++) addCosts[i] = dists[pivot][i];
		Arrays.fill(addFrom, pivot);
		
		for(int i = 0; i < n-1; i++){
			int who = -1;
			double bestCost = Double.POSITIVE_INFINITY;
			for(int j = 0; j < n; j++){
				if(included[j]) continue;
				if(addCosts[j] < bestCost){
					bestCost = addCosts[j];
					who = j;
				}
			}
			included[who] = true;
			outer:
			for(int j = 0; j < n; j++){
				if(included[j]) continue;
				if(squaredTime){
					if(dists[who][addFrom[j]] != -1 && Math.abs(dists[who][addFrom[j]] - addCosts[j]) >= addCosts[j]) continue;
				}else{
					for(int k = 0; k < n; k++){
						if(dists[who][k] != -1 && dists[k][j] != -1 && Math.abs(dists[who][k] - dists[k][j]) >= addCosts[j]) continue outer;
					}
				}
				if(dists[who][j] == -1) dists[who][j] = dists[j][who] = df.getDistance(entries.get(who).obj, entries.get(j).obj);
				if(dists[who][j] < addCosts[j]){
					addCosts[j] = dists[who][j];
					addFrom[j] = who;
				}
			}
		}
		return dists;
	}

	
}
