package bonsai.indexes;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import metrics.core.MetricList;
import metrics.core.MetricObject;
import bonsai.javaimps.BustosSelector;

public class PMTree<T> extends AbstractMTree<T> {

	public PMTree(T[] data, DistanceFunction df, int capacity, int nPivots, boolean bustos){
		super(new PMTreeNode<T>(df, capacity, bustos ? extractPivotsBustos(data, nPivots, df) : extractPivots(data, nPivots)), df);
		for(T o: data){
			if(o != null) insert(o);
		}
	}
	
	protected static <T> List<T> extractPivots(T[] data, int nPivots){
		nPivots = Math.min(nPivots, data.length-1);
		List<T> ret = new ArrayList<T>();
		Random rnd = new Random(124957592);
		for(int i = 0; i < data.length; i++){
			if(rnd.nextInt(data.length-i) < nPivots){
				ret.add(data[i]);
				data[i] = null;
				nPivots--;
			}
		}
		return ret;
	}
	
	@SuppressWarnings("unchecked")
	protected static <T> List<T> extractPivotsBustos(T[] data, int nPivots, DistanceFunction df){
		nPivots = Math.min(nPivots, data.length-1);
		MetricObjectWrapper<T>[] mObs = new MetricObjectWrapper[data.length];
		for(int i = 0; i < data.length; i++){
			mObs[i] = new MetricObjectWrapper<T>(data[i], df);
		}
		MetricListWrapper<MetricObjectWrapper<T>> mList = new MetricListWrapper<MetricObjectWrapper<T>>(mObs, df);
		
		BustosSelector selector = new BustosSelector(nPivots);
		List<MetricObject> pivots = selector.getPivots(mList, nPivots);
		
		List<T> ret = new ArrayList<T>();
		for(MetricObject mo: pivots){
			ret.add(((MetricObjectWrapper<T>)mo).getData());
		}
		for(int i = 0; i < data.length; i++){
			if(ret.contains(data[i])) data[i] = null;
		}
		return ret;
	}
	
	/* Workaround classes for interfacing with Arne's design. */
	private static class MetricObjectWrapper<T> implements MetricObject<T>{
		
		private static final long serialVersionUID = -9055504015842673338L;
		private T data;
		private static int counter;
		private int id;
		private DistanceFunction df;
		
		public MetricObjectWrapper(T data, DistanceFunction df){
			this.data = data;
			this.df = df;
			id = counter++;
		}
		
		@Override
		public T getData() {
			return data;
		}

		@Override
		public double getDistance(MetricObject<T> other) {
			return df.getDistance(data, other.getData());
		}

		@Override
		public Object getUid() {
			return id;
		}

		@Override
		public void setData(T data) {
			this.data = data;
			id = counter++;
		}	
	}
	
	private static class MetricListWrapper<T extends MetricObject<?>> extends AbstractList<T> implements MetricList<T>{

		private static final long serialVersionUID = 4335317077581856435L;
		private T[] data;
		private DistanceFunction df;
		
		public MetricListWrapper(T[] data, DistanceFunction df){
			this.data = data;
			this.df = df;
		}
		
		@Override
		public T get(int index) {
			return data[index];
		}

		@Override
		public int size() {
			return data.length;
		}

		@Override
		public double getDistance(T a, T b) {
			return df.getDistance(a.getData(),b.getData());
		}
	}
}
