package bonsai.javaimps;


import java.io.Serializable;
import metrics.core.*;
import bonsai.javaimps.*;
import java.util.*;

public class LAESA implements Serializable{

	/**
	 * 
	 */
	private static final long serialVersionUID = -3704554245121477018L;
	volatile PivotSelector ps;
	public int numpivots;
	List<MetricObject> pivots;
	double pdistance[][]; 
	MetricList<MetricObject> m;

        private LAESA() {};
        
	public LAESA(MetricList m) {
		this(m,20);
	}
	
	public LAESA(MetricList m,int numpivots) {
		this(m,numpivots,new BustosSelector(numpivots));
	}
	
	public LAESA(MetricList m,int numpivots,PivotSelector ps) {
		this.m = m;	
		this.ps = ps;
		this.numpivots = numpivots;
		index();
	}
	
	public void index() {
		pivots = ps.getPivots(m,numpivots);
		pdistance = new double[pivots.size()][m.size()];
		for(int i = 0;i<m.size();i++) {
			MetricObject mo = m.get(i);
			for(int j = 0;j<pivots.size();j++) {
				pdistance[j][i] = m.getDistance(mo,pivots.get(j));
			}
		}
	}


	public List<MetricObject> range_search(MetricObject q,double r) {
		double dlo[] = new double[m.size()];
		double dhi[] = new double[m.size()];
		Arrays.fill(dhi,Double.MAX_VALUE);
		
		ArrayList<Integer> next = new ArrayList<Integer>(m.size());
		ArrayList<Integer> prev = new ArrayList<Integer>(m.size());
		for(int i = 0;i<m.size();i++) {
			next.add(i);
			prev.add(i);
		}
		
		ArrayList<MetricObject> reslist = new ArrayList<MetricObject>();
		ArrayList<Integer> temp;
		for(int i = 0;i<pivots.size();i++) {
			temp = prev;
			prev = next;
			next = temp;
			next.clear();
			MetricObject p = pivots.get(i);
			double qp = m.getDistance(q,p);
			//System.out.println(posset.size() + " " + qp);
			
			for(Integer j:prev) {
				dlo[j] = Math.max(dlo[j],Math.max(0,Math.abs(qp-pdistance[i][j])));
				dhi[j] = Math.min(dhi[j],qp+pdistance[i][j]);
				if(dhi[j]<r) {
					reslist.add(m.get(j));
				} else if (dlo[j] <= r) {
					next.add(j);
				}
			}
		}
		
		//All that remains:
		for(Integer j:next) {
			if(m.getDistance(m.get(j),q) <=r) {
				reslist.add(m.get(j));
			}
		}
		return reslist;
		
	}

	public PivotSelector getPs() {
		return ps;
	}

	public void setPs(PivotSelector ps) {
		this.ps = ps;
	}

	public int getNumpivots() {
		return numpivots;
	}

	public void setNumpivots(int numpivots) {
		this.numpivots = numpivots;
	}

	
	public List<MetricObject<?>> knn_search(MetricObject q,int n) {
		return knn_search(q,n,Integer.MAX_VALUE);
	}
	
	public List<MetricObject<?>> knn_search(MetricObject q,int n,double maxvalue) {
		
		//List<MetricObject> objects = new ArrayList<MetricObject>(m);
		
		double dlo[] = new double[m.size()];
		double dhi[] = new double[m.size()];
		Arrays.fill(dhi,Double.MAX_VALUE);
		
		class Kncomp implements Comparator<Integer> {
			
			double dhi[];
			
			public Kncomp(double dhi[]) {
				this.dhi = dhi;
			}
			
			
			public int compare(Integer i1, Integer i2) {
				return 0-Double.compare(dhi[i1],dhi[i2]);
			}
		}
		Kncomp kncomp = new Kncomp(dhi); 
		PriorityQueue<Integer> pq = new PriorityQueue<Integer>(n,kncomp);
		
        ArrayList<Integer> next = new ArrayList<Integer>(m.size());
		ArrayList<Integer> prev = new ArrayList<Integer>(m.size());
		
        for(int i = 0;i<m.size();i++) {
			next.add(i);
			prev.add(i);
		}
		
		ArrayList<Integer> temp;
		for(int i = 0;i<pivots.size();i++) {
			temp = prev;
			prev = next;
			next = temp;
			next.clear();
			MetricObject p = pivots.get(i);
			double qp = m.getDistance(q,p);
			//System.out.println(posset.size() + " " + qp);
			System.out.println(prev.size());
			for(Integer j:prev) {
				if(pq.contains(j)) {
					pq.remove(j);
				}
				dlo[j] = Math.max(dlo[j],Math.abs(qp-pdistance[i][j]));
				dhi[j] = Math.min(dhi[j],qp+pdistance[i][j]);
				
				
				
				if(pq.size() < n && !pq.contains(j)) {
					pq.add(j);
				} else if(dhi[j]<dhi[pq.peek()] && !pq.contains(j)) {
					pq.remove();
					pq.add(j);
				}
				if (dlo[j] <= dhi[pq.peek()]) {
					next.add(j);
				}
				
			}
			System.out.println(qp + " " + dhi[pq.peek()]);
		}
		
		//All that remains:
		
		int counts = 0;
		//System.out.println(dhi[pq.peek()]);
		
		class LoComp implements Comparator<Integer> {
			
			double dlo[];
			
			public LoComp(double dlo[]) {
				this.dlo = dlo;
			}
			
			
			public int compare(Integer i1, Integer i2) {
				return Double.compare(dlo[i1],dlo[i2]);
			}
		}
		Collections.sort(next,new LoComp(dlo));
		for(Integer j:next) {
			double r = dhi[pq.peek()];
			System.out.println(dlo[j]);
			if(dlo[j] > r)
				break;
			if(!pq.contains(j)) {
				double d = m.getDistance(m.get(j),q);
				if(d >=r )
					continue;
				dlo[j] = d;
				dhi[j] = d;
				counts++;
				pq.remove();
				pq.add(j);
			}
		}

		//System.out.println(dhi[pq.peek()] + " "+ pq.size()+ " " + BenchObject.counts() + " " + counts + " " + q.distance(m.get(0)));
		//System.out.println(kncomp);
                
		ArrayList<MetricObject<?>> mo = new ArrayList<MetricObject<?>>();
                while(pq.size()>0) {
                	mo.add(m.get(pq.remove()));
		}
                Collections.reverse(mo);
		return mo;
		
	}

}
