package bonsai.javaimps;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Arrays;

import metrics.core.MetricList;
import metrics.core.MetricObject;

class DPair<T> implements Comparable<DPair<T>> {
	double d;
	T o;
	
	DPair(double d,T o) {
		this.d = d;
		this.o = o;
	}

	public int compareTo(DPair<T> arg0) {
		// TODO Auto-generated method stub
		return Double.compare(d, arg0.d);
	}	
	
}

class SANode {
	MetricObject<?> root;
	List<Double> pivspace;
	ArrayList<SANode> pivots;
	double maxrange = 0;
	int maxpivots;
	protected static HashMap<MetricObject,ArrayList<Double>> pivmap;
		
	public SANode(MetricObject root,List<Double> pivspace,int maxpivots) {
		this.root = root;
		this.pivspace = pivspace;
		this.pivspace = pivspace;
		this.maxpivots = maxpivots;
		this.pivots = new ArrayList<SANode>();
	}
	
	public void index(ArrayList<DPair<MetricObject>> data) {
		if(data.size() == 0) {
			pivots = new ArrayList<SANode>();
			return;
		}
		Collections.sort(data);
		
		this.maxrange = data.get(data.size()-1).d;
		
		ArrayList<MetricObject> ndata = new ArrayList<MetricObject>();
		ArrayList<double[]> distmap = new ArrayList<double[]>();
		
		
//		System.out.println("Starting " + root.toString());
		
		for(DPair<MetricObject> dp:data) {
			boolean badobject = false;
			double omap[] = new double[pivots.size()];
			
			for(int i = 0;i<pivots.size();i++) {
				
				double pdist = SATree.data.getDistance(pivots.get(i).root,dp.o);
				omap[i] = pdist;
				//System.out.println(pdist);
				if(pdist < dp.d) {
					badobject = true;
					
				}
			}
			if(badobject) {
				ndata.add(dp.o);
				distmap.add(omap);
			} else {
				//System.out.println(dp.o.toString() + dp.d);
				ArrayList<Double> npivspace = pivmap.get(dp.o);
				int k = npivspace.size() > maxpivots?npivspace.size()-maxpivots:0;
				pivots.add(new SANode(dp.o,npivspace.subList(k,npivspace.size()),maxpivots));
			}
		}
		
		ArrayList<DPair<MetricObject>> pivdata[] = new ArrayList[pivots.size()];
		for(int i = 0;i<pivots.size();i++) {
			pivdata[i] = new ArrayList<DPair<MetricObject>>();
		}
		for(int i = 0;i<ndata.size();i++) {
			
			MetricObject obj = ndata.get(i);
			//System.out.println(obj);
			double dmap[] = distmap.get(i);
			if(dmap.length < pivots.size()) {
				double tmap[] = dmap;
				dmap = new double[pivots.size()];
				System.arraycopy(tmap,0, dmap,0,tmap.length);
				for(int j=tmap.length;j<pivots.size();j++) {
					dmap[j] = SATree.data.getDistance(pivots.get(j).root,obj);
				}
			}
			//Now dmap is filled up
			int minpivot = 0;
			double mindist = dmap[0];
			for(int j =1;j<dmap.length;j++) {
				//System.out.print(dmap[j] + " ");
				
				if(dmap[j] < mindist) {
					minpivot = j;
					mindist = dmap[j];
				}
			}
			//System.out.println();
			pivmap.get(obj).add(mindist);
			pivdata[minpivot].add(new DPair(mindist,obj));
		}
		//ndata.clear();
		//distmap.clear();
		
		for(int i = 0;i<pivots.size();i++) {
			Collections.sort(pivdata[i]);
//			System.out.println(pivots.get(i).root	 + " " + root);
//			for(int j = 0;j<pivdata[i].size();j++) {
//				System.out.print(pivdata[i].get(j).o + " " +pivdata[i].get(j).d + " ");
//			}
//			System.out.println();
//			System.out.println();
//				
			pivots.get(i).index(pivdata[i]);
		}	
	}

	public void range_search(MetricObject query, double range,
			double t, double lastdist, ArrayList<Double> queryspace,ArrayList<MetricObject> retlist) {

		if(lastdist < range)
			retlist.add(root);
		
		List<Double> kqspace;
		if(queryspace.size() > maxpivots) {
			kqspace = queryspace.subList(queryspace.size()-maxpivots,queryspace.size());
		} else {
			kqspace = queryspace;
		}
//		System.out.println("Start " + root.toString() + " " +  kqspace.size());
		
		if(lastdist < this.maxrange +range && pivots != null) {
			ArrayList<SANode> goodneighb = new ArrayList<SANode>();
			ArrayList<Double> gooddists = new ArrayList<Double>();
			double tmind = Double.MAX_VALUE;
			for(int i = 0;i<pivots.size();i++) {
//				System.out.println(pivots.get(i).root + " " + pivots.get(i).pivspace.size());
				
				
				double mindist = 0;
				for(int j = 0;j<kqspace.size();j++) {
					mindist  = Math.max(Math.abs(kqspace.get(j)-pivots.get(i).pivspace.get(j)),mindist);
					
				}
				
				if(mindist <= range + pivots.get(i).maxrange) {
//					System.out.println(mindist);
					goodneighb.add(pivots.get(i));
					double qpdist = SATree.data.getDistance(query,pivots.get(i).root);
					tmind = Math.min(tmind,qpdist);
					gooddists.add(qpdist);
//					System.out.println("Good:" + pivots.get(i).root + " " + qpdist);
//					System.out.println(tmind);
				} else {
					
					//double qpdist = SATree.data.getDistance(query,pivots.get(i).root);
//					System.out.println("Bad:" + pivots.get(i).root + " " + qpdist + " " + mindist);
//					for(int k = 0;k<kqspace.size();k++)
//						System.out.print(" " + kqspace.get(k));
//					System.out.println();
//					for(int k = 0;k<kqspace.size();k++)
//						System.out.print(" " + pivots.get(i).pivspace.get(k));
//					System.out.println();
				}
				
				
			}
			for(int i = 0;i<gooddists.size();i++) {
				double dist = gooddists.get(i);
				if(dist - tmind <= t) {
					SANode pivot = goodneighb.get(i);
					queryspace.add(dist);
					pivot.range_search(query, range, t-(dist-tmind), dist, queryspace, retlist);
					queryspace.remove(queryspace.size()-1);
				}
			}
			
		}
		
	}
	
	
}

public class SATree {

	//MetricList<MetricObject> data;
	SANode root;
	int npivots;
	static MetricList<MetricObject> data;
	

	public SATree(MetricList<MetricObject> data) {
		this(data,15);
	}
		
	public SATree(MetricList<MetricObject> data,int npivots) {
		this.data = data;
		this.root = new SANode(data.get(0),new ArrayList<Double>(),npivots);
		ArrayList<DPair<MetricObject>> pdata = new ArrayList<DPair<MetricObject>>(); 
		for(int i = 1;i<data.size();i++) {
			pdata.add(new DPair(data.getDistance(data.get(0),data.get(i)),data.get(i)));
		}
		Collections.sort(pdata);
		SANode.pivmap = new HashMap();
		for(DPair<MetricObject> obj:pdata) {
			ArrayList<Double> a = new ArrayList<Double>();
			a.add(obj.d);
			SANode.pivmap.put(obj.o,a);
		}
		
		this.root.index(pdata);
		SANode.pivmap.clear();
	}
	
	public List<MetricObject> range_search(MetricObject query,double range) {
		double dist = query.getDistance(root.root);
		ArrayList<MetricObject>retlist = new ArrayList<MetricObject>();
		ArrayList<Double> queryspace = new ArrayList<Double>();
		queryspace.add(dist);
		try {
			root.range_search(query,range,2*range,dist,queryspace,retlist);
		} catch(Exception e) {
			e.printStackTrace();
		}
		return retlist;
 	}
	
	
}
