"""
The LAESA index structure.
"""

import util
import random
import pivotselector
import bisect
from sets import Set as set
__metaclass__ = type

try:
    from javaimps import LAESA
except:
    class LAESA:
        def __init__(self, metriclist,rows=20,pivotsel=pivotselector.RandomSelector):
            self.data = metriclist
            self.dfunc = metriclist.getDistance
            self.rows = min(rows,len(data))
            self.heur_data = {}
            self.pivotsel = pivotsel
            self.set_up_matrix()
    
        def set_up_matrix(self):
            U = self.data
            n = len(U)
            #Choose random pivots (for now)
            self.rowlist = self.pivotsel(self.data,self.dfunc).generate(self.rows)
      #      print len(self.rowlist)
            self.matrix = d = util.DMArray((self.rows,n))
            df = self.dfunc
            rl = self.rowlist
            for x in xrange(self.rows):
                for y in xrange(n):
                    d[x,y] = df(rl[x],self.data[y])
                    
        def knn_search(self,query,knn,maxrange=util.inf):
    
            candidates = range(len(self.data))
            knnlist = [(maxrange,-1)]*knn
            d = self.matrix
            
            U = [util.inf]*len(candidates)
            L = [0]*len(candidates)
            
            for p in xrange(len(self.rowlist)):
                d_qp = self.dfunc(query,self.rowlist[p])
                nextcand = []
                for cand in candidates:
                    uplimit = d_qp+d[p,cand]
                    if uplimit < knnlist[-1][0]:
                        #print cand,uplimit,knnlist[-1][1],knnlist[-1][0]
                        #First remove the old:
                        for l in knnlist:
                            dist,r = l
                            if r == cand:
                                knnlist.remove(l)
                        bisect.insort_left(knnlist,(uplimit,cand))
                        knnlist = knnlist[:knn]
                    
                    U[cand] = min(uplimit,U[cand])
                    L[cand] = max(L[cand],abs(d_qp-d[p,cand]))
                    if L[cand] <= knnlist[-1][0]:
                        
                        nextcand.append(cand)
                cand = nextcand
                
                
            #For those that are left
            for cand in candidates:
                if L[cand] <= knnlist[-1][0]:
                    #print dist
                    dist = self.dfunc(query,self.data[cand])
                    if dist < knnlist[-1][0]:
                        for l in knnlist:
                            throwaway,r = l
                            if r == cand:
                                knnlist.remove(l)
                        bisect.insort_left(knnlist,(dist,cand))
                        knnlist = knnlist[:knn]
            print knnlist
            print str(query)
            print [str(self.data[cand]) for dist,cand in knnlist if cand != -1]
            return [self.data[cand] for dist,cand in knnlist if cand != -1]
        
        def range_search(self,query,radius):
            U  = self.data
            n  = len(U)
            d = self.matrix
            p = 0
            dist = self.dfunc
            result = []
            dmin = min
            dmax = max
            C = set(range(n))
            for p in xrange(len(self.rowlist)):
                d_qp = dist(query,self.rowlist[p])
                F = []
                for u in C:
                    if abs(d_qp+(2*d[p,u])) <= radius:
                        result.append(U[u])
                    elif abs(d_qp-d[p,u]) < radius:
                        F.append(u)
                C = F
            
            for u in C:
                if dist(query,U[u]) <= radius:
                    result.append(U[u])
            return result

USABLE = True
INDEX_NAME = "LAESA"
KNN = True
INDEX_CLASS = LAESA
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Mico:1994,Rico-Juan:2002"
