"""
The MB-Tree index structure.
"""
import random

__metaclass__ = type

class MBTree:
	
	def __init__(self,metric):
		self.dfunc = metric.getDistance
		listdata = []
		for element in metric:
			listdata.append(element)
		self.root = Node(listdata,self.dfunc)
		
	def range_search(self, query, range):
		return self.root.range_search(query, range)
	


class Node:


	def __init__(self, data, dfunc, parentCenter = None):
		
		self.dfunc = dfunc
		self.left = None
		self.right = None

		if(parentCenter is None):
			self.s1 = self.removeRand(data)
			self.is_root = True
		else:
			self.s1 = parentCenter
			self.is_root = False


		self.s2 = self.removeRand(data)
		self.r = self.dfunc(self.s1, self.s2)

		data1 = []
		data2 = []

		for element in data:

			# Calculate distances to each subtree center
			d1 = self.dfunc(self.s1, element)
			d2 = self.dfunc(self.s2, element)

			# Update radius with respect to cluster center s1
			if(d1 > self.r):
				self.r = d1

			# Assign each element to the cluster it is closest to
			if(d1 <= d2):
				data1.append(element)
			else:
				data2.append(element)

		# Create left and right subtrees, if we assigned any elements to them
		if(len(data1) > 0):
			self.left = Node(data1, dfunc, self.s1)
		if(len(data2) > 0):
			self.right = Node(data2, dfunc, self.s2)


	def range_search(self, query, range, centerDist = None):

		# Calculate distance from the cluster center to the search element
		if centerDist is None:
			d = self.dfunc(query, self.s1)
		else:
			d = centerDist
		

		# If center distance - radius of cluster > range, no elements can match
		if(d - self.r > range):
			return []

		# If center distance + radius of cluster <= range, all elements match
		if(d + self.r <= range):
			return [self.s1] + self.allElements()

		# Otherwise we recurse on subtrees to find matches
		ret = []
		if self.left is not None:
			ret += self.left.range_search(query, range, d)
		elif self.dfunc(self.s1, query) <= range:
			ret += [self.s1]

		if self.right is not None:
			ret += self.right.range_search(query, range)
		elif self.dfunc(self.s2, query) <= range:
			ret += [self.s2]
			
		return ret

	def allElements(self):
		ret = [self.s2]

		if self.left is not None:
			ret += self.left.allElements()
		if self.right is not None:
			ret += self.right.allElements()
			
		return ret
						

	def removeRand(self, data):
		remInd = random.randint(0,len(data)-1)
		ret = data[remInd]
		del data[remInd]
		return ret;
		


USABLE = True
INDEX_NAME = "MB-Tree"
INDEX_CLASS = MBTree
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Noltemeier:1992"
