"""
The HC index structure.
"""

# XXX See LC. Code sharing should be possible. (I.e., LC could be a subclass 
# with limited functionality)
# Differences:
# 1. Build a similar structure recursively for the inside-set.
# 2. Use a h-function (h(n)) to control the bucket size for a node containing 
# n objects.
# Could use use a lower limit to bucket size (i.e., if there are at most b 
# objects, don't recurse) to save space -- but that could increase the number 
# of distance comparisons.

# XXX See the experiments by Andersson to figure out the optimal parameters,
# as well as the choice of h(n).

# Each node has a covering radius for the union of the inside and outside
# parts -- so that if d(c,q) > cr'(c) + r, the recursion can stop. This can 
# also be used for automatic inclusion, both for the inside and outside
# parts.

# Fredriksson uses a different pivot selection method from Chavez and Navarro;
# he minimizes the mean squared distance to the rest, using random sampling.

# In Fredriksson:2005 this is called UVPT (unbalanced VP-tree)
# XXX (Or is this really the same?)

__metaclass__ = type



from random import randint
from util import GenericMetricList
from aesa import AESA
from hc import HC
from pivotselector import MeanSquareDistance
from util import GenericMetricList
import util


class PHC(HC):
    def __init__(self,metric):
        self.bucketsize = 10
        self.arity = 5
        data = metric
        self.dfunc = metric.getDistance
        self.index(list(data))
            
    def split(self,list):
        return len(list)
    
   
    def index(self,data):
       
        self.pivots = []
        
        
        for pivot in self.pivots:
            data.remove(pivot)
        
        self.included = {}
        self.maxrad = {}
        r = self.arity+1
        hn = self.split(data)
        
        for i in xrange(self.arity):
            if len(data) == 0:
                break
            msq = MeanSquareDistance(data,self.dfunc)
            #print "Before"
            
            pivot = msq.generate(1)[0]
            #print "After"
            self.pivots.append(pivot)
            data.remove(pivot)
            dlist = [(self.dfunc(pivot,o),o) for o in data]
            util.sort(dlist)
            plist = dlist[:(hn/r)]
            hn -= len(plist)
            r -= 1
            #print len(plist)
            for d,p in plist:
                data.remove(p)
            self.maxrad[pivot] = len(plist) > 0 and max(plist)[0] or 0
            #print pivot,self.maxrad[pivot],len(plist),plist,hn
            #print 
            self.included[pivot] = self.indexlist(plist)
        #print len(data),data
        self.excluded = self.indexlist([(0,o) for o in data])
                
    def range_search(self,query,range):
        retlist = []
        fullyin = False
        #print "Excluded",self.excluded
        for pivot in self.pivots:
            #print pivot
            #print "Included",self.included
            dp = self.dfunc(pivot,query)
            if dp <= range:
                retlist.append(pivot)
            if dp <= self.maxrad[pivot] + range:
                retlist.extend(self.included[pivot].range_search(query,range))
            if dp + range <= self.maxrad[pivot]:
                fullyin = True
                #print str(pivot) + " is fully in"
                break
        if not fullyin:
            retlist.extend(self.excluded.range_search(query,range))
        return retlist
                               
            
USABLE = True
INDEX_NAME = "PHC"
INDEX_CLASS = PHC
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Fredriksson:2005,Fredriksson:2006"