import random
import util
from sets import Set as set
class Selector(object):
    def __init__(self,data,dfunc):
        self.data = data
        self.dfunc = dfunc

class RandomSelector(Selector):
    
    def generate(self,numpivots=20):
        rows = min(len(self.data),numpivots)
        return [self.data[x] for x in util.sample(xrange(len(self.data)),min(len(self.data),numpivots))]

class MeanSquareDistance(Selector):
    
    def generate(self,numpivots=20,randomsample=-1,randompivots=-1):
        if randomsample <= 0:
            randomsample = len(self.data)
        if randompivots <= 0:
            randompivots = numpivots
            
        pivots = []
        if len(self.data) == 1:
            return self.data
        
        while len(pivots) < numpivots:
          
            randpivs = util.sample(self.data,randompivots)
            randsamps = util.sample(self.data,randomsample)
          
            minsq = []
            for p in randpivs:
                pdist = 0
                for o in randsamps:
                    dop = self.dfunc(p,o)
                    pdist += dop*dop
                minsq.append((pdist,p))
            minsq.sort()
          
            for d,p in minsq:
                if p not in pivots:
                    pivots.append(p)
                    break
        return pivots
                
    
class BustosSelector(Selector):
    def generate(self,numpivots=20,randomsample=-1): 
        if randomsample < 0:
            randomsample = len(self.data)
        rows = min(len(self.data),numpivots)
        objects = set(range(rows))
        pivotlist = []
        d = self.dfunc
        while len(pivotlist) < rows:
            #Randomly select candidate pivots
            randomsample = min(len(objects),randomsample)
            candidates = util.sample(objects,min(len(objects),randomsample))
            queries = util.sample(objects,min(len(objects),randomsample))
            tobjects  = util.sample(objects,min(len(objects),randomsample))
            maxsum = 0
            maxcand = 0

            ndists = []
            for i in xrange(len(queries)):
                (q,o) = (self.data[queries[i]],self.data[tobjects[i]])
                ndists.append(0)
                for p in pivotlist:
                    ndists[-1] = max(ndists[-1],abs(d(q,p)-d(p,o)))
            maxsum = 0
            maxcand = candidates[0]
            for candidate in candidates:
                nsum = 0
                p = self.data[candidate]
                for i in xrange(len(queries)):   
                    (q,o) = (self.data[queries[i]],self.data[tobjects[i]])
                    nsum += max(ndists[i],abs(d(q,p)-d(p,o)))
                if nsum > maxsum :
                    maxsum = nsum
                    maxcand = candidate
            objects.remove(maxcand)
            pivotlist.append(self.data[maxcand])
        return pivotlist
'''
class MaxdistSelector(Selector):
    def generate(self,numpivots=20,randomsample=-1):
        if randomsample < 0:
            randomsample = 3*numpivots
        rows = min(len(self.data),numpivots)
        randomsample = min(len(self.data),randomsample)
        objects = set(range(rows))
        p1 = random.randint(0,rows-1)
        pivotlist = [self.data[p1]]
        objects.remove(p1)
        while len(pivotlist) < rows:
            samples = util.sample(objects,min(len(objects),randomsample))
            maxsamp = 0
            maxdist = 0
            for sample in samples:
                sdist = min([self.dfunc(self.data[sample],p) for p in pivotlist])
                if sdist > maxdist:
                    maxsamp = sample
                    maxdist = sdist
            pivotlist.append(self.data[maxsamp])
            objects.remove(maxsamp)
        return pivotlist
'''


