"""
The SA-Tree index structure.

http://citeseer.ist.psu.edu/navarro99searching.html

"""

__metaclass__ = type

import util

nodes = 0
try:
    from bonsai.javaimps import SATree
except Exception,e:
    print e
    class SATree:
        def __init__(self, metriclist,npivots=10):
            
            self.root = SANode(metriclist[0],metriclist.getDistance,npivots,list())
            self.dfunc = metriclist.getDistance
            l = [(x,list()) for i,x in enumerate(metriclist) if i > 0] #
            global nodes
            nodes = 0
            self.root.index(self.root.sortrange(l))
            print "There are %d inner nodes"%nodes
            
        def range_search(self,query,range):
            dist = self.root.dfunc(self.root.center,query)
            return self.root.range_search(query,range,2*range,dist,[dist])
        
        
        def knn_search(self,query,n):
            retlist = util.Heap()
            queue = util.Heap()
            queue.push((0,self.dfunc(query,self.root.center)-self.root.maxdist,self.root))
            #print dir(queue)
            range = util.inf
            while len(queue) > 0:
                (t,d,b) = queue.pop()
                #print t,d,range
                if t > range:
                    break
    
                if d <= range:
                    retlist.push((1.0/(d+b.maxdist),(b.center)))
                    if len(retlist) == n+1:
                        retlist.pop()
                    if len(retlist) == n:
                        (d,c) = retlist.peek()
                        range = 1.0/d
                        #print range
    
                    ndists = [(self.dfunc(c.center,query),c) for c in b.pivots]
                    cdist = len(ndists) > 0 and min(ndists)[0] or 0
                    #print cdist
                    for vdist,v in ndists:
                        #print "1"
                        queue.push((((t+vdist-cdist)/2),vdist-v.maxdist,v))
            retp = []
            while retlist:
                retp.append(retlist.pop()[1])
            return retp
                        
                
    
class SANode:
    def __init__(self,center,dfunc,npivots,pspace):
        self.dfunc = dfunc
        self.center = center
        self.npivots = npivots
        self.pspace = pspace[-self.npivots:]
    
    def sortrange(self,list):
        ranges = []
        for obj,pspace in list:
            dist = self.dfunc(obj,self.center)
            pspace.append(dist)
            ranges.append((dist,obj,pspace))
        util.sort(ranges)
        return ranges
    
    def index(self,ranges):
        
        if len(ranges) == 0:
            self.maxdist = 0
            self.pivots = []
            return
        global nodes
        nodes += 1
        self.maxdist = ranges[-1][0]
        #Find all pivots
        distmap = []
        pivotobjs = []
        for dis,obj,pspace in ranges:
            badobject = False
            objdistmap = []
            for p,_ in pivotobjs:
                pdist = self.dfunc(obj,p)
                objdistmap.append(pdist)
                if pdist < dis:
                    badobject = True
            distmap.append((obj,pspace,objdistmap))
            if not badobject:
                pivotobjs.append((obj,pspace))
        
        pivbuckets = [list() for x in pivotobjs]
        
        #Fill up distmap and find closest 
        for obj,pspace,objdistmap in distmap:
            if (obj,pspace) in pivotobjs:
                continue
            for j in xrange(len(objdistmap),len(pivotobjs)):
                objdistmap.append(self.dfunc(obj,pivotobjs[j][0]))
            
            minbucket = 0
            mindist = util.inf
            for i,d in enumerate(objdistmap):
                if d < mindist:
                    minbucket = i
                    mindist = d
            pspace.append(objdistmap[minbucket])
            pivbuckets[minbucket].append((mindist,obj,pspace))
        self.pivots = [SANode(p,self.dfunc,self.npivots,pspace) for p,pspace in pivotobjs]
        #print self.pivots
        [util.sort(p) for p in pivbuckets]
        #print [len(p) for p in pivbuckets]
        [p.index(pivbuckets[i]) for i,p in enumerate(self.pivots)]
    
    def range_search(self,query,r,t,lastq,queryspace):
        
        retlist = []
        #
        if lastq < r:
            retlist.append(self.center)
        
        if len(queryspace) > self.npivots:
            queryspace = queryspace[len(queryspace)-self.npivots:]
           
        
        if lastq < self.maxdist + r and hasattr(self,'pivots') and len(self.pivots) > 0:
            #print queryspace
            goodneighb = []
            for pivot in self.pivots:
                #print pivot.center,self.dfunc(pivot.center,query),pivot.maxdist,pivot.pspace
                if len(pivot.pspace) == 0 or len(queryspace) == 0 or \
                 max([abs(q-p) for q,p in zip(queryspace,pivot.pspace)]) <= (pivot.maxdist + r):
                    goodneighb.append(pivot)
            
            neighd = [(self.dfunc(x.center,query),x) for x in goodneighb]
            #print neighd
            if len(neighd) > 0:
                mind = min(neighd)[0]
            for dist,b in neighd:
                if dist - mind  <= t:
                    qlist = list(queryspace)
                    qlist.append(dist)
                    retlist.extend(b.range_search(query,r,t-(dist-mind),dist,qlist))
        return retlist

        
KNN = True
USABLE = True
INDEX_NAME = "SA-Tree"
INDEX_CLASS = SATree
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Navarro:1999,Navarro:2002"