"""
SSS index structure.
"""

__metaclass__ = type
import util

class SSSNode:
    def __init__(self,dfunc,center,maxbucket=1,alphalimit=0.4):
        self.dfunc = dfunc
        self.rc = 0
        self.isleaf = True
        self.objectlist = []
        self.pivotlist = []
        self.maxbucket = maxbucket
        self.alphalimit = alphalimit
        self.center = center
    
    def add(self,object):
        centerdist = self.dfunc(object,self.center)
        self.rc = max(self.rc,centerdist)
        
        #If it is less than maxbucket in objectlist, just add it
        if self.isleaf:
            self.objectlist.append(object)
            if len(self.objectlist) > self.maxbucket:
                #Make it an innner node
                self.isleaf = False
                [self.add(o) for o in self.objectlist]
                self.objectlist = []
            return
            
            
        #We know that this is not a leaf
        #First find the closest pivot
        closepivot = None
        closedist = util.inf
        for p in self.pivotlist:
            pdist = self.dfunc(object,p.center)
            if pdist < closedist:
                closepivot = p
                closedist = pdist
             
        if closedist >= 2*self.rc*self.alphalimit:
            #Add pivot
            self.pivotlist.append(SSSNode(self.dfunc,object))
        else:
            closepivot.add(object)
        
    
    def range_search(self,query,radius):
        retlist = []
        #print len(self.pivotlist)
        for p in self.pivotlist:
            #print p.objectlist
            
            pdist = self.dfunc(p.center,query)
            #print pdist,(hasattr(p,'pivotlist') and [o.rc for o in p.pivotlist] or p.objectlist)
            if pdist < radius:
                retlist.append(p.center)
            if pdist - radius <= p.rc:
                if p.isleaf:
                    retlist.extend([x for x in p.objectlist if self.dfunc(x,query) < radius])
                else:
                    retlist.extend(p.range_search(query,radius))
        return retlist

class SSSTree:
    def __init__(self,metric,alphalimit=0.4):
        self.metric = metric
        self.dfunc = metric.getDistance
        #The first object is the center
        self.root = SSSNode(self.dfunc,center=self.metric[0],alphalimit=alphalimit)
        for i in xrange(1,len(metric)):
            self.root.add(metric[i])
        #print len(self.root.pivotlist)
        
        
    def range_search(self,query,radius):
        retlist =  self.root.range_search(query,radius)
        if self.dfunc(query,self.root.center) < radius:
            retlist.append(self.root.center)
        return retlist


USABLE = True
INDEX_NAME = "SSSTree"
INDEX_CLASS = SSSTree
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Brisboa:2008" # XXX