"""
Some basic utilities useful to many index structures.
"""

from __future__ import division
from __future__ import generators
from array import array as Array
import random
from math import floor
#from heapq import *

__metaclass__ = type



#Trying to import inf as safe as possible
#See http://www.python.org/dev/peps/pep-0754/ for problems with inf in Python

inf = 1.0E300
try:
    inf = 1.0E500
except:
    pass

try:
    import numpy
    inf = numpy.inf
except:
    pass

try:
    import java.lang.Double
    inf = java.lang.Double.POSITIVE_INFINITY
except:
    pass

#Should be OK now

#Copied from http://norvig.com/python-iaq.html
import copy
class DefaultDict(dict):
    import copy
    """Dictionary with a default value for unknown keys."""
    def __init__(self, default):
        self.default = default

    def __getitem__(self, key):
        if not key in self: 
            self[key] = copy.deepcopy(self.default)
        return self.get(key)
        


#Python 2.2 does not include random.sample
#The other sample function only needs an iterable    
sample = None
try:
    sample = random.sample
except:
    def sample(list,n):
        ret = []
        count = 0
        for count,l in enumerate(list):
            if count < n:
                ret.append(l)
            else:
                pos = random.randint(0,count)
                if pos < n:
                    ret[pos] = l
        return ret
            
        
#Multiple dimensions array object
#A bit simple atm, and does not have enough functions to truly replace numpy.array
class MulArray(object):
    def __init__(self,type,dimensions):
        self._dimensions = list(dimensions)
        self._factors = Array('i',[0]*len(dimensions))
        product = 1
        for i in range(len(dimensions) - 1,-1,-1):
            self._factors[i] = product
            product *= dimensions[i]
        self._data = Array(type,[0]*product)
        #print self._dimensions

    def getOffset(self,indices):
        try:
            len(indices)
        except:
            indices = [indices]
        if len(indices) != len(self._dimensions):
            raise IndexError
        offset = 0
        for i,dim in enumerate(self._dimensions):
            if indices[i] < 0 or indices[i] >= dim:
                raise IndexError
            offset += self._factors[i] * indices[i]
        return offset
    
    def __getitem__(self,indices):
        return self._data[self.getOffset(indices)]

    def __setitem__(self,indices,value):
        self._data[self.getOffset(indices)] = value

    def __len__(self):
        #print self._dimensions
        return self._dimensions[0]
try:
    import numpy
    DMArray = numpy.zeros
except:
    class DMArray(MulArray):
        def __init__(self,dimensions):
            super(DMArray,self).__init__('d',dimensions)
def argmin(func, seq):
	"""
	Returns an object x in seq for which func(x) yields a minimum value.
	If func has a __getitem__ method (i.e., it is a mapping), func[x] is
	used instead of funx(x).
	"""
	if hasattr(func, '__getitem__'):
		func = func.__getitem__
	it = iter(seq)
	min_obj = it.next()
	min_val = func(min_obj)
	for obj in it:
		val = func(obj)
		if val < min_val:
			min_val = val
			min_obj = obj
	return min_obj

class bunch(dict):
	def __init__(self, *args, **kwds):
		super(bunch, self).__init__(*args, **kwds)
		# XXX:
		# To avoid circular references with the extension classes, we
		# can't just say self.__dict__ = self:
		self.__dict__.update(self)

# XXX Add some form of check, similar to iselement() in ElementTree?
# XXX Have special-purpose attributes such as "key" and "val"?
# XXX Dumping to dot format -- what to use for label keys?
# XXX Iteration
# XXX Why not let the children be a public attribute, like attribs?
# (Current scheme taken from ElementTree.)
class Tree:
		
	"""
	A simple tree structure inspired by ElementTree.
	"""
	
	def __init__(self, attrib=None, **extra):
		if attrib is None: attrib = {}
		attrib.update(extra)
		self.attrib = attrib
		self._children = []
	
	def __len__(self):
		return len(self._children)
	
	def __getitem__(self, index):
		return self._children[index]
	
	def __setitem__(self, index, tree):
		self._children[index] = tree
	
	def __delitem__(self, index):
		del self._children[index]
	
	def __getslice__(self, start, stop):
		return self._children[start:stop]
	
	def __setslice__(self, start, stop, trees):
		for tree in trees:
			tree.parent = self
		self._children[start:stop] = list(trees)
	
	def __delslice__(self, start, stop):
		gone = self._children[start:stop]
		del self._children[start:stop]
	
	def append(self, tree):
		self._children.append(tree)
	
	def insert(self, index, tree):
		self._children.insert(index, tree)
	
	def remove(self, tree):
		self._children.remove(tree)
	
	def clear(self):
		self.attrib.clear()
		self._children = []
	
	def has(self, key):
		return key in self.attrib
		
	def get(self, key, default=None):
		return self.attrib.get(key, default)
	
	def set(self, key, value):
		self.attrib[key] = value
	
	def keys(self):
		return self.attrib.keys()
	
	def items(self):
		return self.attrib.items()
	
	def _get_children(self):
		return self._children
	
	children = property(_get_children)


def SubTree(parent, *args, **kwds):
	"""
	Factory function for creating and adding subtrees.
	"""
	child = Tree(*args, **kwds)
	parent.append(child)
	return child


# XXX The three node types may not really be needed; try to refactor them away?
# XXX Explicitly document the root thing? That the root is consistent with all?
# (It needn't be, really.)
class GiST:
	"""
	A generic tree-based index structure, inspired by the Generic Search
	Tree [1]. It must be supplied with a policy object (see below) as a 
	constructor argument. Note that deletion is currently not implemented.
	
	Each node (subtree) will have a 'node type' (accessed through
	node.get('type')) of either 'internal', 'leaf', or 'object'. Object nodes
	have no children, and serve only to reference an object along with its
	key information. Leaf nodes are the immediate parents of object nodes,
	and internal nodes are all other nodes.
	
	A policy object should support the following methods. In all cases, the 
	tree parameter will be a util.Tree instance, or equivalent.
	
	assign(obj, tree):
	
	Assign the object to the given tree (which might be, for example, an
	empty tree, newly created as a leaf node).
	
	hits(query, tree):
	
	Should return the objects compatible with the query that are directly
	available from the root node of the tree (that is, without any recursive
	search) in the form of an iterable object.
	
	consistent(query, tree):
	
	Should return True if the query object is consistent with the
	information available directly in the subtree root (e.g., a
	key), indicating that further search in the subtree may be
	warranted. It should return False if the subtree may be pruned
	from the search.
		
	combine(tree)
	
	Should combine the key information (bounding predicate) in the tree 
	node based on the information in its subtrees. The node should
	represent the union of its subnodes (as tightly as possible) after 
	this method has been called. It may be assumed that this method has been
	called on each of the subtrees since the last time the subtree was
	modified.
	
	situate(obj, tree):
	
	Should pick (that is, return) the subtree that should receive the given 
	object during insertion. Should adjust the key information (bounding
	predicate) in the returned tree node under the assumption that the
	given object is added to some descendant node.
	
	redistribute(tree, sibling):
	
	Should divide the subtrees of the original tree among it and its newly
	created sibling.
	
	[1] http://gist.cs.berkeley.edu/
	"""
	
	default_capacity = 1024
	
	def __init__(self, policy, capacity=default_capacity):
		self.capacity = capacity
		self.policy = policy
		self.root = Tree(type='leaf')
	
	# XXX adjust() needs the distance info from situate() (used in ajustment).
	# Should, perhaps, situate automatically adjust? The distance must also explicitly
	# be stored when the object is stored in a node... *And* for pivots, in internal nodes.
	def add(self, obj):
		policy = self.policy
		parent = None
		node = self.root
		path = []
		while True:
			path.append((node, parent))
			if node.get('type') == 'leaf': break
			parent = node
			node = policy.situate(obj, node) # XXX Will adjust the key stuff node
		child = SubTree(node, type='object')
		policy.assign(obj, child)
		#new_root = False
		# XXX Needn't really use combine() everywhere here, because situate() does some
		# adjustment
		for node, parent in reversed(path):
			if len(node) > self.capacity:
				if parent is None:
					# Splitting the root:
					parent = Tree(type='internal')
					parent.append(node)
					self.root = parent
					#new_root = True
				sibling = SubTree(parent, type=node.get('type'))
				policy.redistribute(node, sibling)
				policy.combine(sibling)
				policy.combine(node)
			else:
				break
		# XXX The root needn't have key info, does it?
		#if new_root:
		#	policy.combine(self.root)
	
	def search(self, query):
		policy = self.policy
		queue = [self.root]
		while queue:
			node = queue.pop()
			if node.get('type') == 'object':
				for hit in policy.hits(query, node):
					yield hit
			elif policy.consistent(query, node):
				queue.extend(reversed(node))

try:
    import heapq
    class Heap(list):
        def push(self,item):
            heapq.heappush(self,item)
        def pop(self):
            return heapq.heappop(item)
        def peek(self):
            return self[0]
except Exception,e:
    print e
    
try:
    from java.util import PriorityQueue
    from java.lang import Comparable
    
    class Comp(java.lang.Comparable):
        def __init__(self,value):
            self.value = value
        def compareTo(self,other):
            if self.value < other.value:
                return -1
            if self.value == other.value:
                return 0
            else:
                return 1
    
    class Heap(object):
        def __init__(self):
            self.pq = PriorityQueue()
        def push(self,item):
            self.pq.add(Comp(item))
        def pop(self):
            return self.pq.remove().value
        def peek(self):
            return self.pq.peek().value
        def __len__(self):
            return self.pq.size()
except Exception,e:
    print e
    
    


class DistQueue:
	
	"""
	A fixed-capacity priority queue, with the distances from a query
	as priority keys.
	"""
	
	def __init__(self, query, capacity, dfunc):
		self.query = query
		self.capacity = capacity
		self.dfunc = dfunc
		self.contents = []
	
	def add(self, obj, dist=None):
		queue = self.contents
		key = -self.dfunc(self.query, obj)
		entry = (key, obj)
		if len(queue) < self.capacity:
			heappush(queue, entry)
		else:
			current = queue[0]
			if current < entry:
				heapreplace(queue, entry)
	
	def get_radius(self):
		return -self.contents[0][0]
		
	def is_full(self):
		return len(self.contents) == self.capacity

	def get_contents(self, range=None):
		return [o for (k, o) in self.contents if
			range is None or -k < range]

def ball_overlap(r1, r2, dist, outside=False):
	lower_bound = dist - r2
	if outside: lower_bound *= -1
	return r1 >= max(lower_bound, 0)

def median(data):
	"""
	In-place (destructive) median calculation. The parameter is
	assumed to be a list.
	"""
	data.sort() # Usually faster than Python-implemented Randomized-Select
	idx = (len(data)-1)//2
	if not len(data) % 2 and len(data):
		return (data[idx] + data[idx+1])/2.0
	else:
		return data[idx]

def F(p1, p2):
	"""
	Spearman's footrule correlation measure between permutations. See 
	PPIndex.permutation() for a description of permutation format.
	"""
	return sum([abs(x-y) for (x, y) in zip(p1, p2)])

def unoptimized_search(self, query, range, count):
	if range is not None and count is not None:
		return self.knn_search(query, count, range)
	if range is not None:
		return self.range_search(query, range)
	else:
		if count is None:
			count = 1
		return self.knn_search(query, count)




try:
    from bonsai.indexes import DistanceFunction
    
    class JavaDistanceFunction(DistanceFunction):
        
        def __init__(self, df):
            self.df = df
        def getDistance(self, a,b):
            return self.df(a,b)
    
    from bonsai.indexes import CountingDistanceFunction
    
    def getCountingDistanceFunction(df):
        return CountingDistanceFunction(JavaDistanceFunction(df))
    
except:
    pass    

class MetricArray(object):
    def __init__(self,data):
        self.data = data
    
    def __getitem__(self,pos):
        try:
            (i,j) = pos
            return i<j and self.data[i][j] or self.data[j][i]
        except:
            return self.data[pos]
        
    def __setitem__(self,pos,val):
        try:
            (i,j) = pos
            if i<j:
                self.data[i][j] = val
            else:
                self.data[j][i] = val
        except:
            self.data[pos] = val

from metrics import GenericMetricList

def bestDivision(ilist,p,sumweight=0.5):
    
    lastp = 0 
    best = (-len(ilist),0) # (point,value)
    for first in xrange(len(ilist)):
        while lastp < len(ilist) and ilist[lastp] < ilist[first] + 2*p:
            lastp += 1
        goodness = sumweight*(first+(len(ilist)-lastp))-(1-sumweight)*abs(first-(len(ilist)-lastp))
        if goodness >best[0]:
            if lastp < len(ilist):
                best = (goodness,(ilist[first]+ilist[lastp])/2)
            else:
                best = (goodness,(ilist[first]+p))
    print best[1],len(ilist)/2,ilist[int(len(ilist)/2)]    
    return best[1]


class Bucket(object):
    
    def __init__(self,dfunc,objects,lastobjects=100):
        self.bucklist = []
        self.dfunc = dfunc
        self.lastobjects = lastobjects
        self.objects = [(x,p[:lastobjects]) for x,p in objects]
        #print len(self.objects),len(self.objects[0][1])
        
    def range_search(self,query,range,queryspace):
        retlist = []
        queryspace = queryspace[:self.lastobjects]
        #print len(queryspace),len(self.objects)
        for o,pspace in self.objects:
            if len(queryspace) != len(pspace):
                print "Bucket size error",len(queryspace),len(pspace)
            if len(queryspace) > 0 and len(pspace) > 0:
                maxs = min([q+p for q,p in zip(queryspace,pspace)])
                mins = max([abs(q-p) for q,p in zip(queryspace,pspace)])
            else:
                maxs = util.inf
                mins = 0
            #print self.dfunc(o,query),maxs,mins,queryspace,pspace
            if mins <= range and (maxs <= range or self.dfunc(o,query) < range):
                #print self.dfunc(o,query),maxs,mins
                retlist.append(o)
        return retlist

from java.util import Collections
from java.util import Comparator
class CompSort(Comparator):
    def __init__(self):
        pass
    def compare(self,o1,o2):
        if o1 < o2:
            return -1
        elif o1 > o2:
            return 1
        else:
            return 0            
         
comp = CompSort()
def sort(ilist):
     Collections.sort(ilist,comp)