    """
The VP-Tree index structure.
"""

import random
from util import DistQueue, bunch, ball_overlap, median, unoptimized_search

__metaclass__ = type

class VPTree:

    def __init__(self, metric):
        self.data = metric
        self.dfunc = metric.getDistance
        self.root = self.build_tree(list(self.data))

    def select_vp(self, data):
        'Selects a vantage point (simply chosen uniformly at random).'
        return random.choice(data)
    
    def build_tree(self, data):
        if not data: return None
        node = bunch()
        p = node.p = self.select_vp(data)
        d = self.dfunc
        mu = node.mu = median([d(p, s) for s in data])
        left = []
        right = []
        for s in data:
            if s is p: continue
            if d(p, s) < mu:
                left.append(s)
            else:
                right.append(s)
        node.left = self.build_tree(left)
        node.right = self.build_tree(right)
        return node
    
    def range_search(self, query, radius):
        result = []
        d = self.dfunc
        def search(node):
            if node is None: return
            dp = d(query, node.p)
            if dp <= radius: result.append(node.p)
            if ball_overlap(radius, node.mu, dp):
                search(node.left)
            if ball_overlap(radius, node.mu, dp, outside=True):
                search(node.right)
        search(self.root)
        return result
    
    def knn_search(self, query, count, range=None):
        d = self.dfunc
        queue = DistQueue(query, count, d)
        def search(node):
            if node is None: return
            dp = d(query, node.p)
            queue.add(node.p, dp)
            if ball_overlap(queue.get_radius(), node.mu, dp) \
                    or not queue.is_full():
                search(node.left)
            if ball_overlap(queue.get_radius(), node.mu, dp,
                            outside=True) or not queue.is_full():
                search(node.right)
        search(self.root)
        return queue.get_contents(range)

USABLE = True
INDEX_NAME = "VP-Tree"
INDEX_CLASS = VPTree
OPTIMIZES_COMBINED_SEARCH = False
BIBTEX_KEYS = "Uhlmann:1991,Uhlmann:1991a,Yianilos:1993"
