package metrics;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import metrics.core.RandomGeneratedMetric;

public class ClusteredEuclidianGenerator implements RandomGeneratedMetric<Euclidian> {
	
	int dims;
	int clusts;
	final static double dev = 0.5;
	double clusters[][];
	
	public ClusteredEuclidianGenerator(int dimensions,int clusters) {
		dims = dimensions;
		clusts = clusters;
	}
	
	public List<Euclidian> generateRandomObjects(int amount,long seed) {
		Random random = new Random(seed);
		//First, generate clusters, if we haven't already
		if(clusters == null) {
			clusters = new double[clusts][dims];
			for(int i = 0;i<clusts;i++)
				for(int j = 0;j<dims;j++)
					clusters[i][j] = random.nextDouble();
		}
			
		ArrayList<Euclidian> list = new ArrayList<Euclidian>();
		
		for(int i = 0;i<amount;i++) {
			//The points have an equal chance of belonging to a cluster
			//They also have an equal chance of not belonging to any cluster
			int n = random.nextInt(clusts+1);
			double data[] = new double[dims];
			double dir[] = new double[dims];
			if(n==clusts) {
				for(int j = 0;j<dims;j++)
					data[j] = random.nextDouble();
			} else {
				//First we find a random direction
				double dlen = 0.0;
				for(int j = 0;j<dims;j++) {
					dir[j] = random.nextGaussian();
					dlen += dir[j]*dir[j];
				}
				dlen = Math.sqrt(dlen);
				double fact = (random.nextGaussian()*dev)/dlen;
				for(int j = 0;j<dims;j++) {
					data[j] = clusters[n][j] + (dir[j]*fact);
				}
				
			}
			list.add(new Euclidian(data));
			
			
		}

		return list;
	}
}