package metrics;
import java.awt.image.*;
import java.awt.color.*;
import java.io.*;
import javax.imageio.*;



public class ImageMetricHelper {
	
	//White in CIEXYZ:
	//public static double xyzwhite[]={0.9642,1.000,0.8249};
	
	
	
	public static BufferedImage getImage(String filename) throws IOException{
		//System.out.println(filename);
		return javax.imageio.ImageIO.read(new File(filename));
	}
	
	private static float labhelp(float t) {
		if(t > 0.008856)
			return (float) Math.pow(t,1.0/3);
		else
			return (7.787f*t) + (16.0f/116);
		
	}
	/*
	 * Magical function
	 * See Color Science (Wysezecki/Stiles) p. 166 for reference
	 */
	final static ColorSpace cs= new ICC_ColorSpace(ICC_Profile.getInstance(ColorSpace.CS_sRGB));
	final static float A = 0.055f;
	final static float G = 2.4f;
	final static float P = 12.92f;
	
	private static float XYZhelp(float t) {
		if(t <= 0.0405) {
			return t/P;
		} else {
			return (float)Math.pow((t+A)/(1+A),G);
		}

	}
	
	private static float[] RGBtoXYZ(float target[]) {
		
		float mul = 1.0f/256;
		float rlin = XYZhelp(mul*target[0]);
		float glin = XYZhelp(mul*target[1]);
		float blin = XYZhelp(mul*target[2]);
		
		float ret[] = {
				0.4124f*rlin + 0.3576f*glin + 0.1805f*blin,
				0.2126f*rlin + 0.7152f*glin + 0.0722f*blin,
				0.0193f*rlin + 0.1192f*glin + 0.9505f*blin,
				
		};
		return ret;
	}
	
	public static void RGBtoLAB(float target[]) {
		//L:0,a:1,b:2
		//X:0,Y:1,Z:2
		float divs[]= RGBtoXYZ(target);
		if(divs[1] > 0.008867)
			target[0] = (116*(float)Math.pow(divs[1],1.0/3))-16;
		else
			target[0] =  903.3f*divs[1];
		target[1] = (500*(labhelp(divs[0])-labhelp(divs[1])));
		target[2] = (200*(labhelp(divs[1])-labhelp(divs[2])));
	}
	/*
	public static double[][] toCIELAB(BufferedImage bi) {
		//System.out.println("Start CIELAB");
		
		double ret[][] = new double[bi.getWidth()*bi.getHeight()][3];
		int k =0;
		for(int i = 0;i<bi.getWidth();i++) {
			for(int j = 0;j<bi.getHeight();j++,k++) {
				int rgb= bi.getRGB(i,j);
				//float xyz[] = cs.fromCIEXYZ(normcomps);
				RGBtoLAB(bi.getRGB(i,j),ret[k]);
			}
		}
		//System.out.println("Stop CIELAB");
		return ret;
		
	}
	*/
        
        public static float[][] toCIELABXY(BufferedImage bi) {
		//System.out.println("Start CIELAB");
		float ret[][] = new float[bi.getWidth()*bi.getHeight()][5];
		int k =0;
		
		for(int i = 0;i<bi.getWidth();i++) {
			for(int j = 0;j<bi.getHeight();j++,k++) {
				int rgb= bi.getRGB(i,j);
				//float xyz[] = cs.fromCIEXYZ(normcomps);
				ret[k][0] =((rgb >> 16) & 0xff);
				ret[k][1] = (rgb >> 8) & 0xff;
				ret[k][2] = rgb & 0xff;
				//System.arraycopy(cs.toCIEXYZ(ret[k]),0, ret[k],0,3);
				//cs.toRGB(colorvalue)
				//cs.fromRGB(bi.getRGB(x, y));
				RGBtoLAB(ret[k]);
				
                ret[k][3] = (float)(100.0*i)/bi.getWidth();
                ret[k][4] = (float)(100.0*j)/bi.getHeight();
			}
		}
		//System.out.println("Stop CIELAB");
		return ret;
		
	}
        
        
	
	
	public static void main(String args[]) throws IOException{
		BufferedImage bi = getImage("metrics/images/1.jpg");
		toCIELABXY(bi);
	}

}
