package metrics;
import java.awt.image.*;
import java.awt.color.*;
import java.io.*;
import com.sun.image.codec.jpeg.*;



public class JavaImageMetric {
	
	//White in CIEXYZ:
	//public static double xyzwhite[]={0.9642,1.000,0.8249};
	
	
	
	public static BufferedImage getImage(String filename) throws IOException{
		FileInputStream in = new FileInputStream(filename);
		JPEGImageDecoder decoder = JPEGCodec.createJPEGDecoder(in);
		return decoder.decodeAsBufferedImage();
	}
	
	private static double labhelp(double t) {
		if(t > 0.008856)
			return Math.pow(t,1.0/3);
		else
			return (7.787*t) + (16.0/116);
		
	}
	/*
	 * Magical function
	 * See Color Science (Wysezecki/Stiles) p. 166 for reference
	 */
	
	final static double A = 0.055;
	final static double G = 2.4;
	final static double P = 12.92;
	
	private static double XYZhelp(double t) {
		if(t <= 0.0405) {
			return t/P;
		} else {
			return Math.pow((t+A)/(1+A),G);
		}

	}
	
	private static double[] RGBtoXYZ(int rgb) {
		
		double mul = 1.0/256;
		double rlin = XYZhelp(mul*((rgb >> 16) & 0xff));
		double glin = XYZhelp(mul*((rgb >> 8) & 0xff));
		double blin = XYZhelp(mul*(rgb & 0xff));
		
		double ret[] = {
				0.4124*rlin + 0.3576*glin + 0.1805*blin,
				0.2126*rlin + 0.7152*glin + 0.0722*blin,
				0.0193*rlin + 0.1192*glin + 0.9505*blin,
				
		};
		return ret;
	}
	
	public static void RGBtoLAB(int rgb,double target[]) {
		//L:0,a:1,b:2
		//X:0,Y:1,Z:2
		

		
		double divs[]= RGBtoXYZ(rgb);
		if(divs[1] > 0.008867)
			target[0] = (116*Math.pow(divs[1],1.0/3)) -16;
		else
			target[0] =  903.3*divs[1];
		target[1] = (500*(labhelp(divs[0])-labhelp(divs[1])));
		target[2] = (200*(labhelp(divs[1])-labhelp(divs[2])));
	}
	
	public static double[][] toCIELAB(BufferedImage bi) {
		//System.out.println("Start CIELAB");
		float normcomps[] =new float[3];
		double ret[][] = new double[bi.getWidth()*bi.getHeight()][3];
		int k =0;
		for(int i = 0;i<bi.getWidth();i++) {
			for(int j = 0;j<bi.getHeight();j++,k++) {
				int rgb= bi.getRGB(i,j);
				//float xyz[] = cs.fromCIEXYZ(normcomps);
				RGBtoLAB(bi.getRGB(i,j),ret[k]);
			}
		}
		//System.out.println("Stop CIELAB");
		return ret;
		
	}
	
	
	public static void main(String args[]) throws IOException{
		BufferedImage bi = getImage("metrics/images/1.jpg");
		toCIELAB(bi);
	}

}
