package metrics;

import java.io.Serializable;
import java.sql.Time;
import java.util.*;

import metrics.core.*;



public class QuadraticForm implements Serializable{
	
	public final static float alpha = 0.5f;
	
	public float matrix[];
	public float matmax;
	
	private int totsize;
	private float dmax;
	private int dims[];
	private float weights[];
	
	private float mins[];
	private float maxs[];
    //private float weights[];
	
	private long callcount;
	private long calltime;
	
	public void resetCallcount(){
		callcount = 0;
		
	}
	
	public long getCallcount() {
		return callcount;
	}
	
	public class QFData {
		
		public QFData(Map<Integer,Float> data) {
			indices = new int[data.size()];
			values = new float[data.size()];
			int i = 0;
			for(Map.Entry<Integer,Float> item:data.entrySet()) {
				indices[i] = item.getKey();
				values[i] = item.getValue();
				i++;
			}
		}
		public int indices[];
		public float values[];
	}
	
   
        
	public QuadraticForm(int dims[],float mins[],float maxs[],float weights[]) {
		
		//First, lets make a histogram that devides the space up in dims[0]Xdims[1]...
		totsize = 1;
		for(int i=0;i<dims.length;i++) {
			totsize*=dims[i];
		}
		//histogram = new float[totsize];
		this.dims = dims; 
		this.mins = mins;
		this.maxs = maxs;
		this.weights = weights;
		//Lets make a distance matrix, and assume we have enough memory
			
		dmax = 0.0f;
		for(int k=0;k<dims.length;k++)
			dmax += weights[k]*Math.pow(maxs[k]-mins[k],2);
		dmax = (float) Math.sqrt(dmax);
		
		
		//Generate matrix
		this.matrix = new float[totsize];
		for(int i = 0;i<totsize;i++) {
			this.matrix[i] = distanceFunc(0,i);
		}
		
	
	}
	
	private float distanceFunc(int i,int j) {
		float a[] = histToPos(i);
		float b[] = histToPos(j);
		float sum = 0.0f;
		for(int k = 0;k<dims.length;k++) {
			sum += weights[k]*Math.pow(a[k]-b[k],2);
		}
		return 1-(float)Math.sqrt(sum)/dmax;
		
	}
	
	public float getDistance(QFData hista ,QFData histb) {
		
		
		callcount++;
		float retdist = 0;
		int apos,bpos;
		float aval,bval;
		//long before = System.nanoTime();
		
		ArrayList<Integer> indices = new ArrayList<Integer>();
		ArrayList<Float> diffval = new ArrayList<Float>();
		
		for(int i = 0;i<hista.indices.length;i++) {
			indices.add(hista.indices[i]);
			diffval.add(hista.values[i]);
		}
		int dpos = 0;
		for(int i  =0;i<histb.indices.length;i++) {
			while(indices.get(dpos) < histb.indices[i] && dpos < hista.indices.length-1)
				dpos++;
			if(histb.indices[i] == indices.get(dpos)){
				diffval.set(dpos,Math.abs(diffval.get(dpos)-histb.values[i]));
			} else {
				indices.add(histb.indices[i]);
				diffval.add(histb.values[i]);
			}
		}
		
		for(int i = 0;i<indices.size();i++) {
			apos = indices.get(i);
			aval = diffval.get(i);
			for(int j = 0;j<indices.size();j++) {
				bpos = indices.get(j);
				bval = diffval.get(j);
				retdist += aval*bval*matrix[diffpos(apos,bpos)];
				
								
			}
		}
		//System.out.println(System.nanoTime()-before);
		return (float)Math.sqrt(retdist);
	}
	
	
	
	public QFData genNormalizedHist(float pixelList[][]) {
		TreeMap<Integer,Integer> tmap = new TreeMap<Integer,Integer>();
		TreeMap<Integer,Float> hmap = new TreeMap<Integer,Float>();
		float d = 1.0f/pixelList.length;
		for(int i = 0;i<pixelList.length;i++) {
			int hpos = posToHist(pixelList[i]);
			if(!tmap.containsKey(hpos)) {
				tmap.put(hpos,1);
			} else {
				tmap.put(hpos,tmap.get(hpos)+1);
			}
		}
		for(Map.Entry<Integer, Integer> t:tmap.entrySet()) {
			hmap.put(t.getKey(),t.getValue()*d);
		}
		QFData hist = new QFData(hmap);
		return hist;
	}
	
	public int diffpos(int a,int b) {
		int m = 1;
		int ret = 0;
		for(int i = 0;i<dims.length;i++) {
			m *= dims[i];
			int am = a %m;
			int bm = b%m;
			a -= am;
			b -= bm;
			
			ret += Math.abs(am-bm);
		}
		return ret;
	}
	
	public int posToHist(float[] pos) {
		if(pos.length != dims.length)
			return -1;
		int hist = 0;
		int m = 1;
		for(int i = 0;i<pos.length;i++) {
			//Find pos
			int tpos;
			if(pos[i] < mins[i])
				tpos = 0;
			else if(pos[i] >= maxs[i])
				tpos = dims[i] - 1;
			else {
				//We now know mins[i] <= pos[i] < maxs[i]
				tpos =(int) (dims[i]*(pos[i]-mins[i])/(maxs[i]-mins[i]));
			}
			hist += tpos*m;
			m *= dims[i];
		}
		return hist;
	}
	
	
	public float[] histToPos(int hist) {
		if(hist < 0 || hist >= totsize)
			return null;
		int m = 1;
		float ret[] = new float[dims.length];
		for(int i = 0;i<dims.length;i++) {
			float dwid = (maxs[i]-mins[i])/dims[i];
			ret[i] = (dwid*(hist % dims[i])) + mins[i]; 
		}
		return ret;
	}


	public void generateData(int amount) {
		// TODO Auto-generated method stub
		
	}

	public List<float[]> generateRandomObjects(int amount) {
		// TODO Auto-generated method stub
		return null;
	}
	
}
