package metrics;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Random;
import java.util.Stack;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.io.File;
import metrics.core.*;

public class QuadraticFormImageGenerator implements RandomGeneratedMetric<QuadraticFormImageMetric> {
	String folder;
	private int dimensions[];
	private QuadraticForm qf;

	public QuadraticFormImageGenerator(String folder,int dimensions[],float posweight) {
		//Check if folder exists
		File f = new File(folder);
		if (!f.exists()) {
			System.out.println("Folder does not exists");
			return;
		}
		this.folder = folder;
		
		setDimensions(dimensions,posweight);
	}

	public void setDimensions(int dimensions[],float posweight) {
		this.dimensions = dimensions;
		float mins[] = {0,-100,-100,0,0};
		float maxs[] = {100,100,100,100,100};
        float weights[] = {1.0f,1.0f,1.0f,posweight,posweight};         
		qf = new QuadraticForm(dimensions,mins,maxs,weights);
	}
        
    public QuadraticForm getQf() {
        return qf;
    }
    
    class QTask implements Runnable {
    	String f;
    	public boolean running = true;
    	public boolean stopWhenDone = false; 
    	List<QuadraticFormImageMetric> retlist;
    	QuadraticForm qf;
    	
    	public QTask(String f,List<QuadraticFormImageMetric> retlist,QuadraticForm qf) {
    		this.f = f;
    		this.retlist = retlist;
    		this.qf = qf;
    	}
		public void run() {
			
			try {
				QuadraticFormImageMetric qfim = new QuadraticFormImageMetric(f,qf);
				synchronized (retlist) {retlist.add(qfim);}
			} catch(Exception e) {
				System.out.println(e);
			}
		}
    	
    }
        
    
	public List<QuadraticFormImageMetric> generateRandomObjects(int amount,long seed) {
		Stack<File> dirstack = new Stack<File>();
		dirstack.add(new File(folder));
		ArrayList<QuadraticFormImageMetric> retlist = new ArrayList<QuadraticFormImageMetric>();
		ArrayList<Runnable> joblist = new ArrayList<Runnable>();
		int i = 0;
		Random random = new Random(seed);
		while(dirstack.size() > 0) {
			
			//System.out.println(dirstack.peek());
			File f = dirstack.pop();
			if(f.isDirectory()) {
				System.out.println(f.getPath());
				for(String dir:f.list()) {
					dirstack.add(new File(f,dir));
				}
			} else {
				try {
					//System.out.println("\""+f.getPath()+"\"");
					if(!f.getPath().matches(".*\\.((png)|(jpg)|(jpeg))")) continue;
					//retlist.add(new QuadraticFormImageMetric(f.getAbsolutePath(),qf));
					int trand = random.nextInt(i+1);
					if(trand < amount) {
						Runnable runn = new QTask(f.getAbsolutePath(),retlist,qf);
						if(joblist.size() < amount)
							joblist.add(runn);
						else {
							joblist.set(trand,runn);
						}
					}
					i++;
						
				} catch (Exception e) {
					System.err.println(e);
				}
			}
		}
		
		BlockingQueue<Runnable> bqueue = new LinkedBlockingQueue<Runnable>(amount);
		int numprocs = Runtime.getRuntime().availableProcessors();
		ThreadPoolExecutor tpe = new ThreadPoolExecutor(numprocs,numprocs+2,Long.MAX_VALUE,TimeUnit.SECONDS,bqueue);
		
		try {
			for(Runnable runn:joblist)
				tpe.submit(runn);
			tpe.shutdown();
			tpe.awaitTermination(Long.MAX_VALUE, TimeUnit.SECONDS);
		}catch(Exception e) {
			System.out.println(e);
		}
		
		System.out.println(bqueue.size());
		return retlist;
	}

}
