package metrics.core;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.io.Serializable;


public class GenericMetricList<T extends MetricObject<?>>  
	extends ArrayList<T> implements MetricList<T>,Serializable {
	
	/**
	 * serialversion
	 */
	private static final long serialVersionUID = -9128319825579609715L;
	boolean cachedists;
	MetricDistanceCache<T> cache;
	
	RandomGeneratedMetric<T> ranmet;
	static private int distcount =0;
	
    private GenericMetricList() {}
        
	public GenericMetricList(RandomGeneratedMetric<T> ranmet) {
		this(ranmet,true);
	
	}
	
	
	public GenericMetricList(List<T> metrics) {
		this.addAll(metrics);
	}
	
	public GenericMetricList(RandomGeneratedMetric<T> ranmet,boolean cachedists) {
		this.cachedists =cachedists;
		if(cachedists) {
			cache = new MetricDistanceCache<T>();
		}
		
		this.ranmet = ranmet;
	}
	public double getDistance(T a, T b) {
		
		distcount++;
		if(cachedists) {
			return cache.getDistance(a,b);
		}
		else {
			return a.getDistance(b);
		}
	}
	
	public void generateData(int amount) {
		Random random = new Random();
		generateData(amount,random.nextLong());
	}
	
	public void generateData(int amount,long seed) {
		this.clear();
		this.addAll(ranmet.generateRandomObjects(amount,seed));
	}

	
	public List<T> generateRandomObjects(int amount) {
		Random random = new Random();
		return generateRandomObjects(amount,random.nextLong());
	}
	
	public List<T> generateRandomObjects(int amount,long seed) {
		return ranmet.generateRandomObjects(amount,seed);
	}
	
	public void reset() {
		distcount = 0;
	}
	
	public int getDistCount() {
		return distcount;
	}
}
