package metrics.core;

import java.io.Serializable;
import java.util.LinkedHashMap;
import java.util.Map;


class RevPair  implements Serializable{
	Object a,b;
	
	public RevPair(Object a,Object b) {
		this.a = a;
		this.b = b;
	}
	
	public boolean equals(Object p) {
		RevPair pu;
		try {
			pu = (RevPair)p; 
		} catch(Exception e) {
			return false;
		}
		return (this.a.equals(pu.a) && this.b.equals(pu.b)) || (this.b.equals(pu.a) && this.a.equals(pu.b));
		
	}
	public int hashCode() {
		return a.hashCode() + b.hashCode();
	}
}

public class MetricDistanceCache<T extends MetricObject> extends LinkedHashMap<RevPair,Double> implements Serializable{
	
	
	int cache_size;
	
	public MetricDistanceCache() {
		this(1048576);
	}
	
	public MetricDistanceCache(int size) {
		super(size);
		this.cache_size = size;
		
	}
	
	public double getDistance(T a,T b) {
		//System.out.println(a + " " + b);
		RevPair pair = new RevPair(a.getUid(),b.getUid());
		Double dist = this.get(pair);
		if(dist == null) {
            dist = a.getDistance(b);
			this.put(pair,dist);	 
		}
		return dist;
	}
	
	public boolean removeEldestEntry(Map.Entry eldest) {
		return this.size() > cache_size;
	}

}
