#!python
import unittest
from bonsai import linearscan
import metrics
from sets import Set

class ModTest(unittest.TestCase):
    def __init__(self,module,metric,queries,results,range):
        unittest.TestCase.__init__(self)
        self.module = module
        self.metric = metric
        self.queries = queries
        self.results = results
        self.range = range
          
    def runTest(self):
        print "Testing %s" % self.module.INDEX_NAME
        dfunc = self.metric.getDistance
        self.metric.reset()
        mc = self.module.INDEX_CLASS(self.metric)
        print "Indexing: %d" % self.metric.getDistCount()
        self.metric.reset()
        i = 0
        totlenqueries = 0
        nqueries = 0
        #print self.results
        for result,query in zip(self.results,self.queries):
            self.metric.reset()
            mresult = list(mc.range_search(query,self.range))
            mresult.sort()
            totlenqueries += len(mresult)
            nqueries += self.metric.getDistCount()
            if result != mresult:
                print Set(result).difference(Set(mresult)),Set(mresult).difference(Set(result))
            self.assertEqual(result,
                             mresult,
                             "With range %s and query %s, expected %s,got %s"%(self.range,query,[(r,dfunc(r,query)) for r in result],[(r,dfunc(r,query)) for r in mresult]))
            
            i += 1
        print "%d queries: %d (In total %d results)"% (len(self.queries),nqueries,totlenqueries)
    def id(self):
        return self.module.INDEX_NAME

    def shortDescription(self):
        return self.module.INDEX_NAME
    
class KnnModTest(ModTest):
    def runTest(self):
        print "Testing %s" % self.module.INDEX_NAME
        dfunc = self.metric.getDistance
        self.metric.reset()
        mc = self.module.INDEX_CLASS(self.metric)
        print "Indexing: %d" % self.metric.getDistCount()
        self.metric.reset()
        
        i = 0
        knn = self.range
        for query in self.queries:

            mresult = list(mc.knn_search(query,knn))
            mresult.sort()
            self.assertEqual(self.results[i],mresult,"With range %s and query %s, expected %s,got %s"%(self.range,query,[(str(r),dfunc(r,query)) for r in self.results[i]],[(r,dfunc(r,query)) for r in mresult]))
            i += 1

        print "%d queries: %d "% (len(self.queries),self.metric.getDistCount())
class MetricTest(unittest.TestSuite):
    
    from metrics import euclidian
    def __init__(self,modules,metric=euclidian(),size=1000,numqueries=10,range=0.6):
        unittest.TestSuite.__init__(self)
        #self.numqueries = numqueries
        metric.generateData(size)
        
        queries = metric.generateRandomObjects(numqueries)
        dfunc = metric.getDistance
        ls = linearscan.INDEX_CLASS(metric)
        results = []
        
        print "Performing queries for LinearScan"
        for query in queries:
            result = ls.range_search(query,range)
            result.sort()
            print result
            results.append(result)

        for module in modules:
            self.addTest(ModTest(module,metric,queries,results,range))
    
class KnnTest(unittest.TestSuite):
    from metrics import euclidian
    def __init__(self,modules,metric=euclidian(),size=100,numqueries=10,range=10):
        unittest.TestSuite.__init__(self)
        #self.numqueries = numqueries
        metric.generateData(size)
        
        queries = metric.generateRandomObjects(numqueries)
        dfunc = metric.getDistance
        ls = linearscan.INDEX_CLASS(metric)
        results = []
        
        print "Performing queries for LinearScan"
        for query in queries:
            result = ls.knn_search(query,range)
            result.sort()
            results.append(result)

        for module in modules:
            self.addTest(KnnModTest(module,metric,queries,results,range))

if __name__=="__main__":
    
    #Helper function for import, a bit dirty, but hopefully works
    def my_import(name):
        mod = __import__(name)
        components = name.split('.')
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod
    from metrics import *
    import sys
    print sys.argv
    from bonsai import laesa,usablemods,knnmods

#    modules = len(sys.argv) > 1 and [my_import("bonsai.%s"%x) for x in sys.argv[1:]] or usablemods
    myMods = ['pmtree']
    modules =  [my_import("bonsai.%s"%x) for x in myMods] or usablemods

    print sys.argv
    print modules
    

    mt = MetricTest(modules,metric=clusteredEuclidian(),size=1000,range=0.8)

    
    unittest.TextTestRunner(verbosity=3).run(mt)